# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da
import skimage.morphology as sm

from pyxem.signals import Diffraction2D, LazyDiffraction2D
import pyxem.utils._dask as dt
import pyxem.utils._background_subtraction as bt
import pyxem.utils.diffraction as et
import pyxem.utils._pixelated_stem_tools as pst


class TestSignalDimensionGetChunkSliceList:
    @pytest.mark.parametrize(
        "sig_chunks",
        [(10, 10), (5, 10), (5, 10), (5, 5), (20, 10), (20, 20)],
    )
    def test_chunksizes(self, sig_chunks):
        xchunk, ychunk = sig_chunks
        data = da.zeros((20, 20, 20, 20), chunks=(10, 10, ychunk, xchunk))
        chunk_slice_list = dt._get_signal_dimension_chunk_slice_list(data.chunks)
        assert len(data.chunks[-1]) * len(data.chunks[-2]) == len(chunk_slice_list)
        for chunk_slice in chunk_slice_list:
            xsize = chunk_slice[0].stop - chunk_slice[0].start
            ysize = chunk_slice[1].stop - chunk_slice[1].start
            assert xsize == xchunk
            assert ysize == ychunk

    def test_non_square_chunks(self):
        data = da.zeros((2, 2, 20, 20), chunks=(2, 2, 15, 15))
        chunk_slice_list = dt._get_signal_dimension_chunk_slice_list(data.chunks)
        assert chunk_slice_list[0] == (slice(0, 15, None), slice(0, 15, None))
        assert chunk_slice_list[1] == (slice(15, 20, None), slice(0, 15, None))
        assert chunk_slice_list[2] == (slice(0, 15, None), slice(15, 20, None))
        assert chunk_slice_list[3] == (slice(15, 20, None), slice(15, 20, None))

    def test_one_signal_chunk(self):
        data = da.zeros((2, 2, 20, 20), chunks=(1, 1, 20, 20))
        chunk_slice_list = dt._get_signal_dimension_chunk_slice_list(data.chunks)
        assert len(chunk_slice_list) == 1
        assert chunk_slice_list[0] == np.s_[0:20, 0:20]

    def test_rechunk(self):
        data = da.zeros((2, 2, 20, 20), chunks=(1, 1, 20, 20))
        data1 = data.rechunk((2, 2, 10, 10))
        chunk_slice_list = dt._get_signal_dimension_chunk_slice_list(data1.chunks)
        assert len(chunk_slice_list) == 4

    def test_slice_navigation(self):
        data = da.zeros((2, 2, 20, 20), chunks=(1, 1, 20, 20))
        data1 = data[0, 1]
        chunk_slice_list = dt._get_signal_dimension_chunk_slice_list(data1.chunks)
        assert len(chunk_slice_list) == 1
        assert chunk_slice_list[0] == np.s_[0:20, 0:20]


class TestGetSignalDimensionHostChunkSlice:
    @pytest.mark.parametrize(
        "xy, sig_slice, xchunk, ychunk",
        [
            ((0, 0), np.s_[0:10, 0:10], 10, 10),
            ((5, 5), np.s_[0:10, 0:10], 10, 10),
            ((15, 5), np.s_[10:20, 0:10], 10, 10),
            ((5, 15), np.s_[0:10, 10:20], 10, 10),
            ((19, 19), np.s_[10:20, 10:20], 10, 10),
            ((15, 9), np.s_[0:20, 0:20], 20, 20),
            ((17, 16), np.s_[15:20, 15:20], 15, 15),
            ((17, 16), np.s_[15:20, 10:20], 15, 10),
            ((7, 16), np.s_[0:15, 10:20], 15, 10),
            ((25, 2), False, 10, 10),
            ((2, 25), False, 10, 10),
            ((22, 25), False, 10, 10),
            ((-5, 10), False, 10, 10),
            ((5, -1), False, 10, 10),
        ],
    )
    def test_simple(self, xy, sig_slice, xchunk, ychunk):
        x, y = xy
        data = da.zeros((2, 2, 20, 20), chunks=(1, 1, ychunk, xchunk))
        chunk_slice = dt._get_signal_dimension_host_chunk_slice(x, y, data.chunks)
        assert chunk_slice == sig_slice

    @pytest.mark.parametrize(
        "xy, sig_slice",
        [
            ((0, 0), np.s_[0:10, 0:10]),
            ((28, 5), False),
            ((5, 28), np.s_[0:10, 20:30]),
            ((5, 32), False),
            ((12, 18), np.s_[10:20, 10:20]),
        ],
    )
    def test_non_square(self, xy, sig_slice):
        x, y = xy
        data = da.zeros((2, 2, 30, 20), chunks=(1, 1, 10, 10))
        chunk_slice = dt._get_signal_dimension_host_chunk_slice(x, y, data.chunks)
        assert chunk_slice == sig_slice


class TestAlignSingleFrame:
    @pytest.mark.parametrize(
        "shifts", [[0, 0], [1, 0], [-1, 0], [0, -1], [-1, -1], [-2, -3], [-2, -5]]
    )
    def test_simple(self, shifts):
        x_size, y_size = 5, 9
        image = np.zeros((y_size, x_size), dtype=np.uint16)
        x, y = 3, 7
        image[y, x] = 7
        image_shifted = dt._align_single_frame(image, shifts)
        pos = np.s_[y + shifts[1], x + shifts[0]]
        assert image_shifted[pos] == 7
        image_shifted[pos] = 0
        assert not image_shifted.any()

    @pytest.mark.parametrize(
        "shifts,pos",
        [
            [[0.5, 0.5], np.s_[7:9, 3:5]],
            [[0.0, 0.5], np.s_[7:9, 3]],
            [[0.0, -0.5], np.s_[6:8, 3]],
            [[1.0, -1.5], np.s_[5:7, 4]],
        ],
    )
    def test_subpixel_integer_image(self, shifts, pos):
        x_size, y_size = 5, 9
        image = np.zeros((y_size, x_size), dtype=np.uint16)
        x, y = 3, 7
        image[y, x] = 8
        image_shifted = dt._align_single_frame(image, shifts, order=1)
        assert (image_shifted[pos] >= 2).all()
        image_shifted[pos] = 0
        assert not image_shifted.any()

    @pytest.mark.parametrize(
        "shifts,pos",
        [
            [[-1.0, -2.0], np.s_[5, 2]],
            [[-0.5, -2.0], np.s_[5, 2:4]],
            [[-0.5, -2.5], np.s_[4:6, 2:4]],
            [[-0.25, 0.0], np.s_[7, 2:4]],
        ],
    )
    def test_subpixel_float_image(self, shifts, pos):
        x_size, y_size = 5, 9
        image = np.zeros((y_size, x_size), dtype=np.float32)
        x, y = 3, 7
        image[y, x] = 9
        image_shifted = dt._align_single_frame(image, shifts, order=1)
        assert image_shifted[pos].sum() == 9
        image_shifted[pos] = 0
        assert not image_shifted.any()

    @pytest.mark.parametrize("shifts", [[-0.7, -2.7], [1.1, -1.1]])
    def test_not_subpixel_float_image(self, shifts):
        x_size, y_size = 5, 9
        image = np.zeros((y_size, x_size), dtype=np.float32)
        x, y = 3, 7
        image[y, x] = 9
        image_shifted = dt._align_single_frame(image, shifts, order=0)
        pos = np.s_[y + round(shifts[1]), x + round(shifts[0])]
        assert image_shifted[pos] == 9.0
        image_shifted[pos] = 0
        assert not image_shifted.any()


class TestRemoveBadPixels:
    def test_simple(self):
        data = np.ones((20, 30)) * 12
        data[5, 9] = 0
        data[2, 1] = 0
        output = et.remove_bad_pixels(data, data == 0)
        assert (output == 12).all()


class TestTemplateMatchBinaryImage:
    @pytest.mark.parametrize("x, y", [(13, 32), (76, 32), (87, 21), (43, 85)])
    def test_single_frame(self, x, y):
        disk_r = 5
        disk = sm.disk(disk_r)
        data = np.zeros(shape=(100, 100))
        data[y - disk_r : y + disk_r + 1, x - disk_r : x + disk_r + 1] = disk
        match = et.normalize_template_match(data, disk)
        index = np.unravel_index(np.argmax(match), match.shape)
        assert (y, x) == index


class TestBackgroundRemoval:
    def test_median_sub(self):
        footprint = 10
        numpy_array = np.ones((50, 50))
        numpy_array[20:30, 20:30] = 5

        data = bt._subtract_median(numpy_array, footprint=footprint)
        assert data.sum() != numpy_array.sum()
        assert data.shape == numpy_array.shape
        assert data[0, :].all() == 0

    def test_radial_median_sub(self):
        center_x = 25
        center_y = 25
        numpy_array = np.ones((50, 50))
        numpy_array[20:30, 20:30] = 5
        data = bt._subtract_radial_median(
            numpy_array, center_x=center_x, center_y=center_y
        )
        assert data.sum() != numpy_array.sum()
        assert data.shape == numpy_array.shape
        assert data[0, :].all() == 0

    def test_dog_sub(self):
        min_sigma = 10
        numpy_array = np.ones((50, 50))
        numpy_array[20:30, 20:30] = 5
        data = bt._subtract_dog(numpy_array, min_sigma=min_sigma)
        assert data.sum() != numpy_array.sum()
        assert data.shape == numpy_array.shape



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for Cuda."""

from numba import cuda, int32, float32
import numpy as np

try:
    import cupy as cp

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False


def dask_array_to_gpu(dask_array):
    """
    Copy a dask array to the GPU chunk by chunk
    """
    if not CUPY_INSTALLED:
        raise BaseException("cupy is required")
    return dask_array.map_blocks(cp.asarray)


def dask_array_from_gpu(dask_array):
    """
    Copy a dask array from the GPU chunk by chunk
    """
    if not CUPY_INSTALLED:
        raise BaseException("cupy is required")
    return dask_array.map_blocks(cp.asnumpy)


def to_numpy(array):
    """
    Returns the array as an numpy array
    Parameters
    ----------
    array : numpy or cupy array
        Array to determine whether numpy or cupy should be used
    Returns
    -------
    array : numpy.ndarray
    """
    if is_cupy_array(array):
        import cupy as cp

        array = cp.asnumpy(array)
    return array


def get_array_module(array):
    """
    Returns the array module for the given array
    Parameters
    ----------
    array : numpy or cupy array
        Array to determine whether numpy or cupy should be used
    Returns
    -------
    module : module
    """
    module = np
    try:
        import cupy as cp

        if isinstance(array, cp.ndarray):
            module = cp
    except ImportError:
        pass
    return module


def is_cupy_array(array):
    """
    Convenience function to determine if an array is a cupy array

    Parameters
    ----------
    array : array
        The array to determine whether it is a cupy array or not.
    Returns
    -------
    bool
        True if it is cupy array, False otherwise.
    """
    try:
        import cupy as cp

        return isinstance(array, cp.ndarray)
    except ImportError:
        return False


# must be defined outside the kernel
TPB = 256


@cuda.jit
def _correlate_polar_image_to_library_gpu(
    polar_image, sim_r, sim_t, sim_i, correlation, correlation_m
):
    """
    Custom cuda kernel for calculating the correlation for each template
    in a library at all in-plane angles with a polar image

    Parameters
    ----------
    polar_image: 2D numpy.ndarray or cupy.ndarray
        The image in polar coordinates of shape (theta, R)
    sim_r: 2D numpy.ndarray or cupy.ndarray
        The r coordinates of all spots in all templates in the library.
        Spot number is the column, template is the row
    sim_t: 2D numpy.ndarray or cupy.ndarray
        The theta coordinates of all spots in all templates in the library.
        Spot number is the column, template is the row
    sim_i: 2D numpy.ndarray or cupy.ndarray
        The intensity of all spots in all templates in the library.
        Spot number is the column, template is the row
    correlation: 2D numpy.ndarray or cupy.ndarray
        The output correlation matrix of shape (template number, theta), giving
        the correlation of each template at each in-plane angle
    correlation_m: 2D numpy.ndarray or cupy.ndarray
        The output correlation matrix of shape (template number, theta), giving
        the correlation of the mirror of each template at each in-plane angle
    """
    # each cuda thread handles one element
    # thread blocks should be shaped (1, TPB)
    template, shift = cuda.grid(2)

    # don't do anything for grid positions outside template number range
    # we still use threads with shift > correlation.shape[1] to load spot data
    # to shared memory
    if template >= correlation.shape[0]:
        return

    r_shared = cuda.shared.array(shape=(TPB,), dtype=int32)
    t_shared = cuda.shared.array(shape=(TPB,), dtype=int32)
    i_shared = cuda.shared.array(shape=(TPB,), dtype=float32)

    # thread index within block
    ty = cuda.threadIdx.y

    n_shifts = polar_image.shape[0]

    total_spots = sim_r.shape[1]

    # number of iterations necessary to process all the spots
    iters = total_spots // TPB + 1

    min_spot_iter = min((TPB, total_spots))

    tmp = 0.0
    tmp_m = 0.0
    for i in range(iters):
        # fill up the shared arrays with the template
        spot = ty + i * TPB
        if spot < total_spots:
            r_shared[ty] = sim_r[template, spot]
            t_shared[ty] = sim_t[template, spot]
            i_shared[ty] = sim_i[template, spot]
        else:
            r_shared[ty] = 0
            t_shared[ty] = 0
            i_shared[ty] = 0.0

        # wait for all the threads
        cuda.syncthreads()

        if shift < n_shifts:
            # use the shared arrays to update tmp
            for s in range(min_spot_iter):
                r_sp = r_shared[s]
                if r_sp == 0:
                    break
                t_sp = t_shared[s]
                i_sp = i_shared[s]
                t1 = t_sp + shift
                if t1 >= n_shifts:
                    t1 = t1 - n_shifts
                t2 = n_shifts - t_sp + shift
                if t2 >= n_shifts:
                    t2 = t2 - n_shifts
                tmp += polar_image[t1, r_sp] * i_sp
                tmp_m += polar_image[t2, r_sp] * i_sp

        # wait for all the threads
        cuda.syncthreads()

    # don't update out of range
    if shift >= n_shifts:
        return

    correlation[template, shift] = tmp
    correlation_m[template, shift] = tmp_m



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Indexation generator and associated tools."""

import numpy as np
import lmfit
from transforms3d.euler import mat2euler, euler2mat

from diffsims.utils.sim_utils import get_electron_wavelength

from pyxem.signals import VectorMatchingResults
from pyxem.utils.indexation_utils import (
    index_magnitudes,
    match_vectors,
    OrientationResult,
    get_nth_best_solution,
    index_dataset_with_template_rotation,
)
from pyxem.utils.vectors import detector_to_fourier
from pyxem.utils._signals import (
    _select_method_from_method_dict,
    _transfer_navigation_axes,
)


class IndexationGenerator:
    """Generates an indexer for data using a number of methods.

    .. deprecated:: 0.14.0
    `IndexationGenerator` will be removed in pyxem 0.15.0

    Parameters
    ----------
    signal : ElectronDiffraction2D
        The signal of electron diffraction patterns to be indexed.
    diffraction_library : DiffractionLibrary
        The library of simulated diffraction patterns for indexation.

    Returns
    -------
    ValueError : "use AcceleratedIndexationGenerator or VectorIndexationGenerator"
    """

    def __init__(self, signal, diffraction_library):
        raise ValueError(
            "use AcceleratedIndexationGenerator or VectorIndexationGenerator"
        )


class TemplateIndexationGenerator:
    """Generates an indexer for data using a number of methods.

    .. deprecated:: 0.14.0
    `TemplateIndexationGenerator` will be removed in pyxem 0.15.0

    Parameters
    ----------
    signal : ElectronDiffraction2D
        The signal of electron diffraction patterns to be indexed.
    diffraction_library : DiffractionLibrary
        The library of simulated diffraction patterns for indexation.

    Returns
    -------
    ValueError : "use AcceleratedIndexationGenerator or VectorIndexationGenerator"


    """

    def __init__(self, signal, diffraction_library):
        raise ValueError(
            "use AcceleratedIndexationGenerator or VectorIndexationGenerator"
        )


class AcceleratedIndexationGenerator:
    """
    Generates a template-based indexer that also calculates relative
    rotation of templates.

    Parameters
    ----------
    signal : ElectronDiffraction2D
        The signal of electron diffraction patterns to be indexed.
    diffraction_library : DiffractionLibrary
        The library of simulated diffraction patterns for indexation.

    Notes
    -----
    To be used with minimal template libraries whereby the first euler
    angle is 0. It is this angle that is optimized during indexation.
    """

    def __init__(self, signal, diffraction_library):
        # test that the first euler angle is always 0

        for phase in diffraction_library:
            diffraction_library[phase]["orientations"] = np.array(
                diffraction_library[phase]["orientations"]
            )
            if not np.allclose(
                np.sum(diffraction_library[phase]["orientations"][:, 0]), 0
            ):
                raise ValueError(
                    "Invalid diffraction library! Templates must be generated from orientations where "
                    "the first Euler angle is 0"
                )
        self.signal = signal
        self.library = diffraction_library

    def correlate(
        self,
        n_largest=5,
        include_phases=None,
        **kwargs,
    ):
        """
        Correlates the library of simulated diffraction patterns with the
        electron diffraction signal.

        Parameters
        ----------
        n_largest : int, optional
            Number of best solutions to return, in order of descending match
        include_phases : list, optional
            Names of phases in the library to do an indexation for. By default this is
            all phases in the library.
        n_keep : int, optional
            Number of templates to do a full matching on in the second matching step
        frac_keep : float, optional
            Fraction (between 0-1) of templates to do a full matching on. When set
            n_keep will be ignored
        delta_r : float, optional
            The sampling interval of the radial coordinate in pixels
        delta_theta : float, optional
            The sampling interval of the azimuthal coordinate in degrees
        max_r : float, optional
            Maximum radius to consider in pixel units. By default it is the
            distance from the center of the image to a corner
        intensity_transform_function : Callable, optional
            Function to apply to both image and template intensities on an
            element by element basis prior to comparison
        find_direct_beam : bool, optional
            Whether to optimize the direct beam, otherwise the center of the image
            is chosen
        direct_beam_positions : 2-tuple of floats or 3D numpy array of shape (scan_x, scan_y, 2), optional
            (x, y) coordinates of the direct beam in pixel units. Overrides other
            settings for finding the direct beam
        normalize_images : bool, optional
            normalize the images to calculate the correlation coefficient
        normalize_templates : bool, optional
            normalize the templates to calculate the correlation coefficient
        parallelize_polar_conversion : bool, optional
            use multiple workers for converting the dataset to polar coordinates. Overhead
            could make this slower on some hardware and for some datasets.
        chunks : string or 4-tuple, optional
            internally the work is done on dask datasets and this parameter determines
            the chunking. If set to None then no re-chunking will happen if the dataset
            was loaded lazily. If set to "auto" then dask attempts to find the optimal
            chunk size. If None, no changes will be made to the chunking.
        parallel_workers: int, optional
            the number of workers to use in parallel. If set to "auto", the number
            will be determined from os.cpu_count()

        Returns
        -------
        result : dict

        Notes
        -----
        Internally, this code is compiled to LLVM machine code, so stack traces are often hard to follow on failure. As such it is
        important to be careful with your parameters selection.
        """
        result = index_dataset_with_template_rotation(
            self.signal, self.library, phases=include_phases, n_best=n_largest, **kwargs
        )
        return result


class ProfileIndexationGenerator:
    """Generates an indexer for data using a number of methods.

    Parameters
    ----------
    profile : ElectronDiffraction1D
        The signal of diffraction profiles to be indexed.
    library : ProfileSimulation
        The simulated profile data.

    """

    def __init__(self, magnitudes, simulation, mapping=True):
        self.map = mapping
        self.magnitudes = magnitudes
        self.simulation = simulation

    def index_peaks(self, tolerance=0.1, *args, **kwargs):
        """Assigns hkl indices to peaks in the diffraction profile.

        Parameters
        ----------
        tolerance : float
            The n orientations with the highest correlation values are returned.
        keys : list
            If more than one phase present in library it is recommended that
            these are submitted. This allows a mapping from the number to the
            phase.  For example, keys = ['si','ga'] will have an output with 0
            for 'si' and 1 for 'ga'.
        *args : arguments
            Arguments passed to the map() function.
        **kwargs : arguments
            Keyword arguments passed to the map() function.

        Returns
        -------
        matching_results : ProfileIndexation

        """
        return index_magnitudes(np.array(self.magnitudes), self.simulation, tolerance)


def _refine_best_orientations(
    single_match_result,
    vectors,
    library,
    accelarating_voltage,
    camera_length,
    n_best=5,
    rank=0,
    index_error_tol=0.2,
    method="leastsq",
    vary_angles=True,
    vary_center=False,
    vary_scale=False,
    verbose=False,
):
    """
    Refine a single orientation agains the given cartesian vector coordinates.

    Parameters
    ----------
    single_match_result : VectorMatchingResults
        Pool of solutions from the vector matching algorithm
    n_best : int
        Refine the best `n` orientations starting from `rank`.
        With `n_best=0` (default), all orientations are refined.
    rank : int
        The rank of the solution to start from.
    solution : list
        np.array containing the initial orientation
    vectors : DiffractionVectors
        DiffractionVectors to be indexed.
    structure_library : :obj:`diffsims:StructureLibrary` Object
        Dictionary of structures and associated orientations for which
        electron diffraction is to be simulated.
    index_error_tol : float
        Max allowed error in peak indexation for classifying it as indexed,
        calculated as :math:`|hkl_calculated - round(hkl_calculated)|`.
    method : str
        Minimization algorithm to use, choose from:
        'leastsq', 'nelder', 'powell', 'cobyla', 'least-squares'.
        See `lmfit` documentation (https://lmfit.github.io/lmfit-py/fitting.html)
        for more information.
    vary_angles : bool,
        Free the euler angles (rotation matrix) during the refinement.
    vary_center : bool
        Free the center of the diffraction pattern (beam center) during the refinement.
    vary_scale : bool
        Free the scale (i.e. pixel size) of the diffraction vectors during refinement.

    Returns
    -------
    result : OrientationResult
        Container for the orientation refinement results
    """
    if not isinstance(single_match_result[0], tuple):  # pragma: no cover
        single_match_result = single_match_result[0]
    n_matches = len(single_match_result)

    if n_best == 0:
        n_best = n_matches - rank

    n_best = min(n_matches, n_best)

    top_matches = np.empty(n_best, dtype="object")
    res_rhkls = []

    for i in range(rank, rank + n_best):
        if verbose:  # pragma: no cover
            print(f"# {i}/{n_best} ({n_matches})")

        solution = get_nth_best_solution(single_match_result, "vector", rank=i)

        result = _refine_orientation(
            solution,
            vectors,
            library,
            accelarating_voltage=accelarating_voltage,
            camera_length=camera_length,
            index_error_tol=index_error_tol,
            method=method,
            vary_angles=vary_angles,
            vary_center=vary_center,
            vary_scale=vary_scale,
            verbose=verbose,
        )

        top_matches[i] = result[0]
        res_rhkls.append(result[1])

    res = np.empty(2, dtype=object)
    res[0] = top_matches
    res[1] = np.asarray(res_rhkls)
    return res


def _refine_orientation(
    solution,
    k_xy,
    structure_library,
    accelarating_voltage,
    camera_length,
    index_error_tol=0.2,
    method="leastsq",
    vary_angles=True,
    vary_center=False,
    vary_scale=False,
    verbose=False,
):
    """
    Refine a single orientation agains the given cartesian vector coordinates.

    Parameters
    ----------
    solution : OrientationResult
        Namedtuple containing the starting orientation
    k_xy : DiffractionVectors
        DiffractionVectors (x,y pixel format) to be indexed.
    structure_library : :obj:`diffsims:StructureLibrary` Object
        Dictionary of structures and associated orientations for which
        electron diffraction is to be simulated.
    index_error_tol : float
        Max allowed error in peak indexation for classifying it as indexed,
        calculated as :math:`|hkl_calculated - round(hkl_calculated)|`.
    method : str
        Minimization algorithm to use, choose from:
        'leastsq', 'nelder', 'powell', 'cobyla', 'least-squares'.
        See `lmfit` documentation (https://lmfit.github.io/lmfit-py/fitting.html)
        for more information.
    vary_angles : bool,
        Free the euler angles (rotation matrix) during the refinement.
    vary_center : bool
        Free the center of the diffraction pattern (beam center) during the refinement.
    vary_scale : bool
        Free the scale (i.e. pixel size) of the diffraction vectors during refinement.
    verbose : bool
        Be more verbose

    Returns
    -------
    result : OrientationResult
        Container for the orientation refinement results
    """

    # prepare reciprocal_lattice
    structure = structure_library.structures[solution.phase_index]
    lattice_recip = structure.lattice.reciprocal()

    def objfunc(params, k_xy, lattice_recip, wavelength, camera_length):
        cx = params["center_x"].value
        cy = params["center_y"].value
        ai = params["ai"].value
        aj = params["aj"].value
        ak = params["ak"].value
        scale = params["scale"].value

        rotmat = euler2mat(ai, aj, ak)

        k_xy = k_xy + np.array((cx, cy)) * scale
        cart = detector_to_fourier(k_xy, wavelength, camera_length)

        intermediate = cart.dot(rotmat.T)  # Must use the transpose here
        hklss = lattice_recip.fractional(intermediate) * scale

        rhklss = np.rint(hklss)
        ehklss = np.abs(hklss - rhklss)

        return ehklss

    ai, aj, ak = mat2euler(solution.rotation_matrix)

    params = lmfit.Parameters()
    params.add("center_x", value=solution.center_x, vary=vary_center)
    params.add("center_y", value=solution.center_y, vary=vary_center)
    params.add("ai", value=ai, vary=vary_angles)
    params.add("aj", value=aj, vary=vary_angles)
    params.add("ak", value=ak, vary=vary_angles)
    params.add("scale", value=solution.scale, vary=vary_scale, min=0.8, max=1.2)

    wavelength = get_electron_wavelength(accelarating_voltage)
    camera_length = camera_length * 1e10
    args = k_xy, lattice_recip, wavelength, camera_length

    res = lmfit.minimize(objfunc, params, args=args, method=method)

    if verbose:  # pragma: no cover
        lmfit.report_fit(res)

    p = res.params

    ai, aj, ak = p["ai"].value, p["aj"].value, p["ak"].value
    scale = p["scale"].value
    center_x = params["center_x"].value
    center_y = params["center_y"].value

    rotation_matrix = euler2mat(ai, aj, ak)

    k_xy = k_xy + np.array((center_x, center_y)) * scale
    cart = detector_to_fourier(k_xy, wavelength=wavelength, camera_length=camera_length)

    intermediate = cart.dot(rotation_matrix.T)  # Must use the transpose here
    hklss = lattice_recip.fractional(intermediate)

    rhklss = np.rint(hklss)

    error_hkls = res.residual.reshape(-1, 3)
    error_mean = np.mean(error_hkls)

    valid_peak_mask = np.max(error_hkls, axis=-1) < index_error_tol
    valid_peak_count = np.count_nonzero(valid_peak_mask, axis=-1)

    num_peaks = len(k_xy)

    match_rate = (valid_peak_count * (1 / num_peaks)) if num_peaks else 0

    orientation = OrientationResult(
        phase_index=solution.phase_index,
        rotation_matrix=rotation_matrix,
        match_rate=match_rate,
        error_hkls=error_hkls,
        total_error=error_mean,
        scale=scale,
        center_x=center_x,
        center_y=center_y,
    )

    res = np.empty(2, dtype=object)
    res[0] = orientation
    res[1] = rhklss

    return res


class VectorIndexationGenerator:
    """Generates an indexer for DiffractionVectors using a number of methods.

    Attributes
    ----------
    vectors : DiffractionVectors
        DiffractionVectors to be indexed.
    vector_library : DiffractionVectorLibrary
        Library of theoretical diffraction vector magnitudes and inter-vector
        angles for indexation.

    Parameters
    ----------
    vectors : DiffractionVectors
        DiffractionVectors to be indexed.
    vector_library : DiffractionVectorLibrary
        Library of theoretical diffraction vector magnitudes and inter-vector
        angles for indexation.
    """

    def __init__(self, vectors, vector_library):
        if vectors.cartesian is None:
            raise ValueError(
                "Cartesian coordinates are required in order to index "
                "diffraction vectors. Use the calculate_cartesian_coordinates "
                "method of DiffractionVectors to obtain these."
            )
        else:
            self.vectors = vectors
            self.library = vector_library

    def index_vectors(
        self,
        mag_tol,
        angle_tol,
        index_error_tol,
        n_peaks_to_index,
        n_best,
        *args,
        **kwargs,
    ):
        """Assigns hkl indices to diffraction vectors.

        Parameters
        ----------
        mag_tol : float
            The maximum absolute error in diffraction vector magnitude, in units
            of reciprocal Angstroms, allowed for indexation.
        angle_tol : float
            The maximum absolute error in inter-vector angle, in units of
            degrees, allowed for indexation.
        index_error_tol : float
            Max allowed error in peak indexation for classifying it as indexed,
            calculated as :math:`|hkl_calculated - round(hkl_calculated)|`.
        n_peaks_to_index : int
            The maximum number of peak to index.
        n_best : int
            The maximum number of good solutions to be retained.
        *args : arguments
            Arguments passed to the map() function.
        **kwargs : arguments
            Keyword arguments passed to the map() function.

        Returns
        -------
        indexation_results : VectorMatchingResults
            Navigation axes of the diffraction vectors signal containing vector
            indexation results for each probe position.
        """
        vectors = self.vectors
        library = self.library

        matched = vectors.cartesian.map(
            match_vectors,
            library=library,
            mag_tol=mag_tol,
            angle_tol=np.deg2rad(angle_tol),
            index_error_tol=index_error_tol,
            n_peaks_to_index=n_peaks_to_index,
            n_best=n_best,
            inplace=False,
            *args,
            **kwargs,
        )
        indexation = matched.isig[0]
        rhkls = matched.isig[1].data

        indexation_results = VectorMatchingResults(indexation)
        indexation_results.vectors = vectors
        indexation_results.hkls = rhkls
        indexation_results = _transfer_navigation_axes(
            indexation_results, vectors.cartesian
        )

        vectors.hkls = rhkls

        return indexation_results

    def refine_best_orientation(
        self,
        orientations,
        accelarating_voltage,
        camera_length,
        rank=0,
        index_error_tol=0.2,
        vary_angles=True,
        vary_center=False,
        vary_scale=False,
        method="leastsq",
    ):
        """Refines the best orientation and assigns hkl indices to diffraction vectors.

        Parameters
        ----------
        rank : int
            The rank of the solution, i.e. rank=2 refines the third best solution
        index_error_tol : float
            Max allowed error in peak indexation for classifying it as indexed,
            calculated as :math:`|hkl_calculated - round(hkl_calculated)|`.
        method : str
            Minimization algorithm to use, choose from:
            'leastsq', 'nelder', 'powell', 'cobyla', 'least-squares'.
            See `lmfit` documentation (https://lmfit.github.io/lmfit-py/fitting.html)
            for more information.
        vary_angles : bool,
            Free the euler angles (rotation matrix) during the refinement.
        vary_center : bool
            Free the center of the diffraction pattern (beam center) during the refinement.
        vary_scale : bool
            Free the scale (i.e. pixel size) of the diffraction vectors during refinement.

        Returns
        -------
        indexation_results : VectorMatchingResults
            Navigation axes of the diffraction vectors signal containing vector
            indexation results for each probe position.
        """

        return self.refine_n_best_orientations(
            orientations,
            accelarating_voltage=accelarating_voltage,
            camera_length=camera_length,
            n_best=1,
            rank=rank,
            index_error_tol=index_error_tol,
            method=method,
            vary_angles=vary_angles,
            vary_center=vary_center,
            vary_scale=vary_scale,
        )

    def refine_n_best_orientations(
        self,
        orientations,
        accelarating_voltage,
        camera_length,
        n_best=0,
        rank=0,
        index_error_tol=0.2,
        vary_angles=True,
        vary_center=False,
        vary_scale=False,
        method="leastsq",
    ):
        """Refines the best orientation and assigns hkl indices to diffraction vectors.

        Parameters
        ----------
        orientations : VectorMatchingResults
            List of orientations to refine, must be an instance of `VectorMatchingResults`.
        accelerating_voltage : float
            The acceleration voltage with which the data was acquired.
        camera_length : float
            The camera length in meters.
        n_best : int
            Refine the best `n` orientations starting from `rank`.
            With `n_best=0` (default), all orientations are refined.
        rank : int
            The rank of the solution to start from.
        index_error_tol : float
            Max allowed error in peak indexation for classifying it as indexed,
            calculated as :math:`|hkl_calculated - round(hkl_calculated)|`.
        method : str
            Minimization algorithm to use, choose from:
            'leastsq', 'nelder', 'powell', 'cobyla', 'least-squares'.
            See `lmfit` documentation (https://lmfit.github.io/lmfit-py/fitting.html)
            for more information.
        vary_angles : bool,
            Free the euler angles (rotation matrix) during the refinement.
        vary_center : bool
            Free the center of the diffraction pattern (beam center) during the refinement.
        vary_scale : bool
            Free the scale (i.e. pixel size) of the diffraction vectors during refinement.

        Returns
        -------
        indexation_results : VectorMatchingResults
            Navigation axes of the diffraction vectors signal containing vector
            indexation results for each probe position.
        """
        vectors = self.vectors
        library = self.library

        matched = orientations.map(
            _refine_best_orientations,
            vectors=vectors,
            library=library,
            accelarating_voltage=accelarating_voltage,
            camera_length=camera_length,
            n_best=n_best,
            rank=rank,
            method="leastsq",
            verbose=False,
            vary_angles=vary_angles,
            vary_center=vary_center,
            vary_scale=vary_scale,
            inplace=False,
            parallel=False,
        )

        indexation = matched.isig[0]
        rhkls = matched.isig[1].data

        indexation_results = VectorMatchingResults(indexation)
        indexation_results.vectors = vectors
        indexation_results.hkls = rhkls
        indexation_results = _transfer_navigation_axes(
            indexation_results, vectors.cartesian
        )

        return indexation_results



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for operating on 2D Diffraction Patterns."""

import itertools
from copy import deepcopy
import math

import numpy as np
from scipy.spatial.distance import cdist
from scipy.spatial import ConvexHull
from transforms3d.axangles import axangle2mat

__all__ = [
    "detector_to_fourier",
    "calculate_norms",
    "calculate_norms_ragged",
    "filter_vectors_ragged",
    "filter_vectors_edge_ragged",
    "normalize_or_zero",
    "get_rotation_matrix_between_vectors",
    "get_npeaks",
    "get_angle_cartesian_vec",
    "get_angle_cartesian",
    "filter_vectors_near_basis",
    "vectors_to_polar",
    "to_cart_three_angles",
    "polar_to_cartesian",
    "get_vectors_mesh",
    "get_angles",
    "get_filtered_combinations",
    "convert_to_markers",
    "points_to_polygon",
    "points_to_poly_collection",
    "column_mean",
    "vectors2image",
    "get_three_angles",
]


def detector_to_fourier(k_xy, wavelength, camera_length):
    """Maps two-dimensional Cartesian coordinates in the detector plane to
    three-dimensional coordinates in reciprocal space, with origo in [000].

    The detector uses a left-handed coordinate system, while the reciprocal
    space uses a right-handed coordinate system.

    Parameters
    ----------
    k_xy : numpy.ndarray
        Cartesian coordinates in detector plane, in reciprocal Ångström.
    wavelength : float
        Electron wavelength in Ångström.
    camera_length : float
        Camera length in metres.

    Returns
    -------
    k : numpy.ndarray
        Array of Cartesian coordinates in reciprocal space relative to [000].

    """

    if k_xy.shape == (1,) and k_xy.dtype == "object":
        # From ragged array
        k_xy = k_xy

    # The calibrated positions of the diffraction spots are already the x and y
    # coordinates of the k vector on the Ewald sphere. The radius is given by
    # the wavelength. k_z is calculated courtesy of Pythagoras, then offset by
    # the Ewald sphere radius.

    k_z = np.sqrt(1 / (wavelength**2) - np.sum(k_xy**2, axis=1)) - 1 / wavelength

    # Stack the xy-vector and the z vector to get the full k
    k = np.hstack((k_xy, k_z[:, np.newaxis]))
    return k


def calculate_norms(z):
    """Calculates the norm of an array of cartesian vectors. For use with map().

    Parameters
    ----------
    z : numpy.ndarray
        Array of cartesian vectors.

    Returns
    -------
    norms : numpy.ndarray
        Array of vector norms.
    """
    return np.linalg.norm(z, axis=1)


def calculate_norms_ragged(z):
    """Calculates the norm of an array of cartesian vectors. For use with map()
    when applied to a ragged array.

    Parameters
    ----------
    z : numpy.ndarray
        Array of cartesian vectors.

    Returns
    -------
    norms : numpy.ndarray
        Array of vector norms.
    """
    norms = []
    for i in z:
        norms.append(np.linalg.norm(i))
    return np.asarray(norms)


def filter_vectors_ragged(z, min_magnitude, max_magnitude, columns=[0, 1]):
    """Filters the diffraction vectors to accept only those with magnitudes
    within a user specified range.

    Parameters
    ----------
    min_magnitude : float
        Minimum allowed vector magnitude.
    max_magnitude : float
        Maximum allowed vector magnitude.

    Returns
    -------
    filtered_vectors : numpy.ndarray
        Diffraction vectors within allowed magnitude tolerances.
    """
    # Calculate norms
    norms = np.linalg.norm(z[:, columns], axis=1)
    # Filter based on norms
    norms[norms < min_magnitude] = 0
    norms[norms > max_magnitude] = 0
    filtered_vectors = z[np.where(norms)]

    return filtered_vectors


def filter_vectors_edge_ragged(z, x_threshold, y_threshold):
    """Filters the diffraction vectors to accept only those not within a user
    specified proximity to detector edge.

    Parameters
    ----------
    x_threshold : float
        Maximum x-coordinate in calibrated units.
    y_threshold : float
        Maximum y-coordinate in calibrated units.

    Returns
    -------
    filtered_vectors : numpy.ndarray
        Diffraction vectors within allowed tolerances.
    """
    # Filter x / y coordinates
    z[np.absolute(z.T[0]) > x_threshold] = 0
    z[np.absolute(z.T[1]) > y_threshold] = 0
    filtered_vectors = z[np.where(z.T[0])]

    return filtered_vectors


def normalize_or_zero(v):
    """Normalize `v`, or return the vector directly if it has zero length.

    Parameters
    ----------
    v : numpy.ndarray
        Single vector or array of vectors to be normalized.
    """
    norms = np.linalg.norm(v, axis=-1)
    nonzero_mask = norms > 0
    if np.any(nonzero_mask):
        v[nonzero_mask] /= norms[nonzero_mask].reshape(-1, 1)


def get_rotation_matrix_between_vectors(from_v1, from_v2, to_v1, to_v2):
    """Calculates the rotation matrix from one pair of vectors to the other.
    Handles multiple to-vectors from a single from-vector.

    Find `R` such that `v_to = R @ v_from`.

    Parameters
    ----------
    from_v1, from_v2 : numpy.ndarray
        Vector to rotate _from_.
    to_v1, to_v2 : numpy.ndarray
        Nx3 array of vectors to rotate _to_.

    Returns
    -------
    R : numpy.ndarray
        Nx3x3 list of rotation matrices between the vector pairs.
    """
    # Find normals to rotate around
    plane_normal_from = np.cross(from_v1, from_v2, axis=-1)
    plane_normal_to = np.cross(to_v1, to_v2, axis=-1)
    plane_common_axes = np.cross(plane_normal_from, plane_normal_to, axis=-1)

    # Try to remove normals from degenerate to-planes by replacing them with
    # the rotation axes between from and to vectors.
    to_degenerate = np.isclose(np.sum(np.abs(plane_normal_to), axis=-1), 0.0)
    plane_normal_to[to_degenerate] = np.cross(from_v1, to_v1[to_degenerate], axis=-1)
    to_degenerate = np.isclose(np.sum(np.abs(plane_normal_to), axis=-1), 0.0)
    plane_normal_to[to_degenerate] = np.cross(from_v2, to_v2[to_degenerate], axis=-1)

    # Normalize the axes used for rotation
    normalize_or_zero(plane_normal_to)
    normalize_or_zero(plane_common_axes)

    # Create rotation from-plane -> to-plane
    common_valid = ~np.isclose(np.sum(np.abs(plane_common_axes), axis=-1), 0.0)
    angles = get_angle_cartesian_vec(
        np.broadcast_to(plane_normal_from, plane_normal_to.shape), plane_normal_to
    )
    R1 = np.empty((angles.shape[0], 3, 3))
    if np.any(common_valid):
        R1[common_valid] = np.array(
            [
                axangle2mat(axis, angle, is_normalized=True)
                for axis, angle in zip(
                    plane_common_axes[common_valid], angles[common_valid]
                )
            ]
        )
    R1[~common_valid] = np.identity(3)

    # Rotate from-plane into to-plane
    rot_from_v1 = np.matmul(R1, from_v1)
    rot_from_v2 = np.matmul(R1, from_v2)

    # Create rotation in the now common plane

    # Find the average angle
    angle1 = get_angle_cartesian_vec(rot_from_v1, to_v1)
    angle2 = get_angle_cartesian_vec(rot_from_v2, to_v2)
    angles = 0.5 * (angle1 + angle2)
    # Negate angles where the rotation where the rotation axis points the
    # opposite way of the to-plane normal. Einsum gives list of dot
    # products.
    neg_angle_mask = (
        np.einsum("ij,ij->i", np.cross(rot_from_v1, to_v1, axis=-1), plane_normal_to)
        < 0
    )
    np.negative(angles, out=angles, where=neg_angle_mask)

    # To-plane normal still the same
    R2 = np.array(
        [
            axangle2mat(axis, angle, is_normalized=True)
            for axis, angle in zip(plane_normal_to, angles)
        ]
    )

    # Total rotation is the combination of to plane R1 and in plane R2
    R = np.matmul(R2, R1)

    return R


def get_npeaks(found_peaks):
    """Returns the number of entries in a list. For use with map().

    Parameters
    ----------
    found_peaks : numpy.ndarray
        Array of found peaks.

    Returns
    -------
    len : int
        The number of peaks in the array.
    """
    return len(found_peaks)


def get_angle_cartesian_vec(a, b):
    """Compute the angles between two lists of vectors in a cartesian
    coordinate system.

    Parameters
    ----------
    a, b : numpy.ndarray
        The two lists of directions to compute the angle between in Nx3 float
        arrays.

    Returns
    -------
    angles : numpy.ndarray
        List of angles between `a` and `b` in radians.
    """
    if a.shape != b.shape:
        raise ValueError(
            "The shape of a {} and b {} must be the same.".format(a.shape, b.shape)
        )

    denom = np.linalg.norm(a, axis=-1) * np.linalg.norm(b, axis=-1)
    denom_nonzero = denom != 0.0
    angles = np.zeros(a.shape[0])
    angles[denom_nonzero] = np.arccos(
        np.clip(
            np.sum(a[denom_nonzero] * b[denom_nonzero], axis=-1) / denom[denom_nonzero],
            -1.0,
            1.0,
        )
    ).ravel()
    return angles


def get_angle_cartesian(a, b):
    """Compute the angle between two vectors in a cartesian coordinate system.

    Parameters
    ----------
    a, b : array-like with 3 floats
        The two directions to compute the angle between.

    Returns
    -------
    angle : float
        Angle between `a` and `b` in radians.
    """
    denom = np.linalg.norm(a) * np.linalg.norm(b)
    if denom == 0:
        return 0.0
    return math.acos(max(-1.0, min(1.0, np.dot(a, b) / denom)))


def filter_vectors_near_basis(vectors, basis, columns=[0, 1], distance=None):
    """
    Filter an array of vectors to only the list of closest vectors
    to some set of basis vectors.  Only vectors within some `distance`
    are considered.  If no vector is within the `distance` np.nan is
    returned for that vector.

    Parameters
    ----------
    vectors: array-like
        A two dimensional array of vectors where each row identifies a new vector

    basis: array-like
        A two dimensional array of vectors where each row identifies a vector.

    columns: list
        A list of columns to consider when comparing vectors.  The default is [0,1]

    Returns
    -------
    closest_vectors: array-like
        An array of vectors which are the closest to the basis considered.
    """
    new_vectors = vectors[:, columns]
    if basis.ndim == 1:  # possible bug in hyperspy map with iterating basis
        basis = basis.reshape(1, -1)
    basis = basis[:, columns]
    if len(new_vectors) == 0:
        end_shape = (basis.shape[0], vectors.shape[1])
        new_vectors = np.empty(end_shape)
        new_vectors[:, :] = np.nan
        return new_vectors
    distance_mat = cdist(new_vectors, basis)
    closest_index = np.argmin(distance_mat, axis=0)
    min_distance = distance_mat[closest_index, np.arange(len(basis), dtype=int)]
    closest_vectors = vectors[closest_index]
    if distance is not None:
        if closest_vectors.dtype == int:
            closest_vectors = closest_vectors.astype(float)
        closest_vectors[min_distance > distance, :] = np.nan
    return closest_vectors


def _reverse_pos(peaks, ind=2):
    """Reverses the position of the peaks in the signal.

    This is useful for plotting the peaks on top of a hyperspy signal as the returned peaks are in
    reverse order to the hyperspy markers. Only points up to the index are reversed.

    Parameters
    ----------
    peaks : numpy.ndarray
        Array of peaks to be reversed.
    ind : int
        The index of the position to be reversed.

    """
    new_data = np.empty(peaks.shape[:-1] + (2,))
    for i in range(ind):
        new_data[..., (-i - 1)] = peaks[..., i]
    return new_data


def cluster(
    data, method, columns, column_scale_factors, min_vectors=None, remove_nan=True
):
    vectors = data[:, columns]
    if remove_nan:
        isnan = ~np.isnan(vectors).any(axis=1)
        vectors = vectors[isnan]
    vectors = vectors / np.array(column_scale_factors)
    clusters = method.fit(vectors)
    labels = clusters.labels_
    if min_vectors is not None:
        label, counts = np.unique(labels, return_counts=True)
        below_min_v = label[counts < min_vectors]
        labels[np.isin(labels, below_min_v)] = -1
    one_neg = np.array((-1,) * len(data))[
        :,
        np.newaxis,
    ]
    vectors_and_labels = np.hstack([data, one_neg])
    if remove_nan:
        vectors_and_labels[isnan, -1] = labels
    return vectors_and_labels


def only_signal_axes(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        if self.has_navigation_axis:
            raise ValueError(
                "This function is not supported for signals with a navigation axis"
            )

        return func(*args, **kwargs)

    return wrapper


def vectors_to_polar(vectors, columns=None):
    """Converts a list of vectors to polar coordinates.

    Parameters
    ----------
    vectors : numpy.ndarray
        Array of vectors.
    columns:
        The x and y columns to be used to calculate the
        polar vector.


    Returns
    -------
    polar_vectors : numpy.ndarray
        Array of vectors in polar coordinates.
    """
    if columns is None:
        columns = [0, 1]
    polar_vectors = np.empty(vectors.shape)
    polar_vectors[:, 0] = np.linalg.norm(vectors[:, columns], axis=1)
    polar_vectors[:, 1] = np.arctan2(vectors[:, columns[1]], vectors[:, columns[0]])
    polar_vectors[:, 2:] = vectors[:, 2:]
    return polar_vectors


def to_cart_three_angles(vectors):
    k = vectors[:, 0]
    delta_phi = vectors[:, 1]
    min_angle = vectors[:, 2]
    angles = np.repeat(min_angle, 3)
    angles[1::3] += delta_phi
    angles[2::3] += delta_phi * 2
    k = np.repeat(k, 3)

    return np.vstack([k * np.cos(angles), k * np.sin(angles)]).T


def polar_to_cartesian(vectors):
    k = vectors[:, 0]
    phi = vectors[:, 1]
    return np.vstack([k * np.cos(phi), k * np.sin(phi)]).T


def get_vectors_mesh(g1_norm, g2_norm, g_norm_max, angle=0.0, shear=0.0):
    """
    Calculate vectors coordinates of a mesh defined by a norm, a rotation and
    a shear component.

    Parameters
    ----------
    g1_norm, g2_norm : float
        The norm of the two vectors of the mesh.
    g_norm_max : float
        The maximum value for the norm of each vector.
    angle : float, optional
        The rotation of the mesh in degree.
    shear : float, optional
        The shear of the mesh. It must be in the interval [0, 1].
        The default is 0.0.

    Returns
    -------
    numpy.ndarray
        x and y coordinates of the vectors of the mesh

    """

    def rotation_matrix(angle):
        return np.array(
            [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
        )

    def shear_matrix(shear):
        return np.array([[1.0, shear], [0.0, 1.0]])

    if shear < 0 or shear > 1:
        raise ValueError("The `shear` value must be in the interval [0, 1].")

    order1 = int(np.ceil(g_norm_max / g1_norm))
    order2 = int(np.ceil(g_norm_max / g2_norm))
    order = max(order1, order2)

    x = np.arange(-g1_norm * order, g1_norm * (order + 1), g1_norm)
    y = np.arange(-g2_norm * order, g2_norm * (order + 1), g2_norm)

    xx, yy = np.meshgrid(x, y)
    vectors = np.stack(np.meshgrid(x, y)).reshape((2, (2 * order + 1) ** 2))

    transformation = rotation_matrix(np.radians(angle)) @ shear_matrix(shear)

    vectors = transformation @ vectors
    norm = np.linalg.norm(vectors, axis=0)

    return vectors[:, norm <= g_norm_max].T


##################################################
# Symmetry Methods for Analyzing Glassy Materials
##################################################


def get_angles(angles):
    """This function takes a list of angles and returns the angles between each pair of angles.
    This is useful for finding the angles between three vectors.

    Parameters
    ----------
    angles: numpy.ndarray
        An array of angles in radians.  This is a 2D array with shape (n, 3) where n is the number
        of combinations and 3 is specific combination of angles to determine the difference between.

    """
    all_angles = np.abs(np.triu(np.subtract.outer(angles, angles)))
    all_angles = all_angles[all_angles != 0]
    all_angles[all_angles > np.pi] = np.pi - np.abs(
        all_angles[all_angles > np.pi] - np.pi
    )
    return all_angles


def get_filtered_combinations(
    pks,
    num,
    radial_index=0,
    angle_index=1,
    intensity_index=None,
    intensity_threshold=None,
    min_angle=None,
    min_k=None,
):
    """
    Creates combinations of `num` peaks but forces at least one of the combinations to have
    an intensity higher than the `intensity_threshold`.
    This filter is useful for finding high intensity features but not losing lower intensity
    paired features which contribute to symmetry etc.

    Parameters
    ----------
    pks : numpy.ndarray
        The diffraction vectors to be analyzed
    num : int
        The number of peaks to be combined
    radial_index : int, optional
        The index of the radial component of the diffraction vectors, by default 0
    angle_index : int, optional
        The index of the angular component of the diffraction vectors, by default 1
    intensity_index : int, optional
        The index of the intensity component of the diffraction vectors, by default None
    intensity_threshold : float, optional
        The minimum intensity of the diffraction vectors to be considered, by default None
    min_angle: float, optional
        The minimum angle between two diffraction vectors. This ignores small angles which are
        likely to be from the same feature or unphysical.
    min_k : float, optional
        The minimum difference between the radial component of the diffraction vectors to be
        considered from the same feature, by default 0.05
    """
    if intensity_threshold is not None and intensity_index is not None:
        intensity = pks[:, intensity_index]
        intensity_bool = intensity > intensity_threshold
        pks = pks[intensity_bool]
        intensity = intensity[intensity_bool]
    else:
        intensity = np.ones(len(pks))
    angles = pks[:, angle_index]
    k = pks[:, radial_index]

    angle_combos = np.array(list(itertools.combinations(angles, num)))
    k_combos = np.array(list(itertools.combinations(k, num)))
    intensity_combos = np.array(list(itertools.combinations(intensity, num)))

    if len(angle_combos) == 0:
        angle_combos = np.zeros(shape=(0, 3))
    # Filtering out combinations where there are two peaks close to each other
    if min_angle is not None:
        in_range = (
            np.abs(angle_combos[:, :, np.newaxis] - angle_combos[:, np.newaxis, :])
            < min_angle
        )

        above_angle = np.any(
            np.sum(
                in_range,
                axis=1,
            )
            < 2,
            axis=1,
        )  # See if there are two angles that are too close to each other

    else:
        above_angle = True
    # Filtering out combinations of diffraction vectors at different values for k.
    # This could be faster if we sort the values by k first and then only get the combinations
    # of the values within a certain range.
    if min_k is not None and len(k_combos) > 0:
        mean_k = np.mean(k_combos, axis=1)
        abs_k = np.abs(np.subtract(k_combos, mean_k[:, np.newaxis]))
        in_k_range = np.mean(abs_k, axis=1) < min_k
    else:
        in_k_range = True
    in_combos = above_angle * in_k_range
    if len(angle_combos) == 0:
        combos = angle_combos
        combos_k = []
        combo_inten = []
    elif np.all(in_combos):
        combos = angle_combos
        combos_k = np.mean(k_combos, axis=1)
        combo_inten = np.mean(intensity_combos, axis=1)
    else:
        combos = angle_combos[in_combos]
        combos_k = np.mean(k_combos[in_combos], axis=1)
        combo_inten = np.mean(intensity_combos[in_combos], axis=1)
    return combos, combos_k, combo_inten


def get_three_angles(
    pks,
    k_index=0,
    angle_index=1,
    intensity_index=2,
    intensity_threshold=None,
    accept_threshold=0.05,
    min_k=0.05,
    min_angle=None,
):
    """
    This function takes the angle between three points and determines the angle between them,
    returning the angle if it is repeated using the `accept_threshold` to measure the acceptable
    difference between angle a and angle b.

    Parameters
    ----------
    pks : numpy.ndarray
        The diffraction vectors to be analyzed
    k_index : int, optional
        The index of the radial component of the diffraction vectors, by default 0
    angle_index : int, optional
        The index of the angular component of the diffraction vectors, by default 1
    intensity_index : int, optional
        The index of the intensity component of the diffraction vectors, by default 2
    intensity_threshold : float, optional
        The minimum intensity of the diffraction vectors to be considered, by default None
    accept_threshold : float, optional
        The maximum difference between angle a and angle b to be considered the same angle,
        by default 0.05
    min_k : float, optional
        The minimum difference between the radial component of the diffraction vectors to be
        considered from the same feature, by default 0.05
    min_angle: float, optional
        The minimum angle between two diffraction vectors. This ignores small angles which are
        likely to be from the same feature or unphysical.

    Returns
    -------
    three_angles : numpy.ndarray
        An array of angles between three diffraction vectors.  The columns are:
        [k, delta phi, min-angle, intensity, reduced-angle]
    """
    three_angles = []
    combos, combo_k, combo_inten = get_filtered_combinations(
        pks,
        3,
        radial_index=k_index,
        angle_index=angle_index,
        intensity_index=intensity_index,
        intensity_threshold=intensity_threshold,
        min_angle=min_angle,
        min_k=min_k,
    )
    for c, k, inten in zip(combos, combo_k, combo_inten):
        angular_seperations = get_angles(c)
        min_ind = np.argmin(angular_seperations)
        min_sep = angular_seperations[min_ind]
        angular_seperations = np.delete(angular_seperations, min_ind)
        in_range = np.abs((angular_seperations - min_sep)) < accept_threshold
        is_symetric = np.any(in_range)
        if is_symetric:
            # take the average of the two smaller angles
            avg_sep = np.mean((np.array(angular_seperations)[in_range][0], min_sep))
            min_angle = np.min(c)
            num_times = np.round(min_angle / min_sep)
            three_angles.append(
                [
                    k,
                    avg_sep,
                    min_angle,
                    inten,
                    np.abs(min_angle - (num_times * min_sep)),
                ]
            )
    if len(three_angles) == 0:
        three_angles = np.empty((0, 5))
    return np.array(three_angles)


# =============================================
# Functions for use with the map_vectors method
# =============================================


def column_mean(vectors, columns):
    """Calculate the mean of the columns of a set of vectors. Useful for calculating the mean
    using the map_vectors method.

    Parameters
    ----------
    vectors: numpy.ndarray
        The vectors to be used to calculate the mean
    columns:
        The columns to be used to calculate the mean.
    """
    return np.mean(vectors[:, columns], axis=0)


##############################
# Plotting Diffraction Vectors
##############################


def vectors2image(
    vectors,
    image_size,
    scales,
    offsets,
    indexes=None,
):
    """Convert a set of vectors to an image by binning the vectors into a 2D histogram.

    Parameters
    ----------
    vectors: numpy.ndarray
        The vectors to be binned
    image_size  : tuple
        The size of the image to be produced
    scales: tuple
        The scales of the image to be produced
    offsets: tuple
        The offsets of the image to be produced
    indexes: tuple
        The indexes of the vectors to be used to create the image.
        If None, the first two columns are used.

    Returns
    -------

    """
    if indexes is None:
        indexes = [0, 1]
    red_points = vectors[:, indexes]
    image_range = (
        (offsets[0], offsets[0] + image_size[0] * scales[0]),
        (offsets[1], offsets[1] + image_size[1] * scales[1]),
    )
    im, _, _ = np.histogram2d(
        red_points[:, 0], red_points[:, 1], bins=image_size, range=image_range
    )
    return im


def points_to_poly_collection(points, hull_index=(0, 1)):
    """Convert a set of points to a polygon collection by creating a polygon. The method takes the
    points given and finds the outermost points to create the polygon.

    Parameters
    ----------
    points: numpy.ndarray
        The points to be used to create the polygon (N x 2)
    hull_index:
        The index of the points to be used to create the polygon. The default is (0, 1) which
        means that the first two columns of the points are used to create the polygon.
    """
    try:
        hull = ConvexHull(points[:, hull_index][:, ::-1])
    except:
        return np.array([[0, 0], [0, 0], [0, 0]])
    return hull.points[hull.vertices]


def points_to_polygon(points, num_points=50):
    """Convert a set of points to a polygon by creating a polygon. The method takes the points
    given and finds the outermost points to create the polygon. The number of points in the
    polygon defines the resolution of the polygon.

    Parameters
    ----------
    points: numpy.ndarray
        The points to be used to create the polygon (N x 2)
    num_points
        The number of points on each side (top, bottom, left, right) of the polygon
        used to create the edges of the polygon.

    Returns
    -------
    numpy.ndarray
        The vertices of the polygon

    """
    if len(points) == 0:
        return np.array([[0, 0], [0, 0], [0, 0]])
    sorted_points = points[np.argsort(points[:, 1])]
    min_x = sorted_points[0, 1]
    max_x = sorted_points[-1, 1]
    sorted_index = np.linspace(min_x, max_x, num_points)

    lo = np.searchsorted(sorted_points[:, 1], sorted_index[:-1], side="left")
    hi = np.searchsorted(sorted_points[:, 1], sorted_index[1:], side="right")

    min_points = []
    for l, h, x in zip(lo, hi, sorted_index):
        if l == h:
            pass
        else:
            min_points.append(
                [
                    np.min(sorted_points[l:h][:, 0]),
                    x,
                ]
            )

    max_points = []
    for l, h, x in zip(lo, hi, sorted_index):
        if l == h:
            pass
        else:
            max_points.append([np.max(sorted_points[l:h][:, 0]), x])
    all_points = min_points + max_points[::-1]
    all_points = np.array(all_points)
    return all_points


def convert_to_markers(
    peaks,
    signal,
):
    """Convert a set of (flattened) peaks to a set of markers for plotting. Note that the
    function below only works for 4D signals.


    Parameters
    ----------
    peaks: DiffractionVectors2D
        The peaks to be converted to markers
    signal: pyxem.signals.Signal2D
        The 4-D signal to plot the peaks on

    Returns
    -------

    """
    new_peaks = deepcopy(peaks.data)
    x_axis, y_axis = signal.axes_manager.navigation_axes
    new_peaks[:, 0] = np.round((new_peaks[:, 0] - x_axis.offset) / x_axis.scale)
    new_peaks[:, 1] = np.round((new_peaks[:, 1] - y_axis.offset) / y_axis.scale)
    ind = np.lexsort((new_peaks[:, 1], new_peaks[:, 0]))
    sorted_peaks = new_peaks[ind]
    shape = signal.axes_manager._navigation_shape_in_array
    by_ind_peaks = np.empty(shape, dtype=object)
    by_ind_colors = np.empty(shape, dtype=object)
    num_labels = np.max(new_peaks[:, -1])
    colors_by_index = (
        np.random.random((int(num_labels + 1), 3)) * 0.9
    )  # (Stay away from white)
    colors_by_index = np.vstack((colors_by_index, [1, 1, 1]))
    low_x_ind = np.searchsorted(sorted_peaks[:, 0], range(0, shape[1]), side="left")
    high_x_ind = np.searchsorted(
        sorted_peaks[:, 0], range(1, shape[1] + 1), side="left"
    )
    for i, (lo_x, hi_x) in enumerate(zip(low_x_ind, high_x_ind)):
        x_inds = sorted_peaks[lo_x:hi_x]
        low_y_ind = np.searchsorted(x_inds[:, 1], range(0, shape[0]), side="left")
        high_y_ind = np.searchsorted(x_inds[:, 1], range(1, shape[0] + 1), side="left")
        for j, (lo_y, hi_y) in enumerate(zip(low_y_ind, high_y_ind)):
            x_values = x_inds[lo_y:hi_y, 2]
            y_values = x_inds[lo_y:hi_y, 3]
            labels = np.array(x_inds[lo_y:hi_y, -1], dtype=int)
            by_ind_peaks[j, i] = np.stack((y_values, x_values), axis=1)
            by_ind_colors[j, i] = colors_by_index[labels]
    return by_ind_peaks, by_ind_colors, colors_by_index



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import dask.array as da
import pytest
from matplotlib import pyplot as plt

import hyperspy.api as hs

from pyxem.signals import Diffraction1D, LazyDiffraction1D


class TestComputeAndAsLazy1D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyDiffraction1D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert isinstance(s, Diffraction1D)
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_3d_data_compute(self):
        dask_array = da.random.random((4, 10, 15), chunks=(1, 10, 15))
        s = LazyDiffraction1D(dask_array)
        s.compute()
        assert isinstance(s, Diffraction1D)
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = Diffraction1D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyDiffraction1D)
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_3d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = Diffraction1D(data)
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyDiffraction1D)
        assert data.shape == s_lazy.data.shape


class TestDecomposition:
    def test_decomposition_is_performed(self, electron_diffraction1d):
        s = Diffraction1D(electron_diffraction1d)
        s.decomposition()
        assert s.learning_results is not None

    def test_decomposition_class_assignment(self, electron_diffraction1d):
        s = Diffraction1D(electron_diffraction1d)
        s.decomposition()
        assert isinstance(s, Diffraction1D)


class TestVirtualImaging:
    # Tests that virtual imaging runs without failure

    def test_plot_integrated_intensity(self, electron_diffraction1d):
        roi = hs.roi.SpanROI(left=1.0, right=2.0)
        plt.ion()  # to make plotting non-blocking
        electron_diffraction1d.plot_integrated_intensity(roi)
        plt.close("all")

    def test_get_integrated_intensity(self, electron_diffraction1d):
        roi = hs.roi.SpanROI(left=1.0, right=2.0)
        vi = electron_diffraction1d.get_integrated_intensity(roi)
        assert vi.data.shape == (2, 2)
        assert vi.axes_manager.signal_dimension == 2
        assert vi.axes_manager.navigation_dimension == 0

    @pytest.mark.parametrize("out_signal_axes", [None, (0, 1), (1, 2), ("x", "y")])
    def test_get_integrated_intensity_stack(
        self, electron_diffraction1d, out_signal_axes
    ):
        s = hs.stack([electron_diffraction1d] * 3)
        s.axes_manager.navigation_axes[0].name = "x"
        s.axes_manager.navigation_axes[1].name = "y"

        roi = hs.roi.SpanROI(left=1.0, right=2.0)
        vi = s.get_integrated_intensity(roi, out_signal_axes)
        assert vi.axes_manager.signal_dimension == 2
        assert vi.axes_manager.navigation_dimension == 1
        if out_signal_axes == (1, 2):
            assert vi.data.shape == (2, 3, 2)
            assert vi.axes_manager.navigation_size == 2
            assert vi.axes_manager.signal_shape == (2, 3)
        else:
            assert vi.data.shape == (3, 2, 2)
            assert vi.axes_manager.navigation_size == 3
            assert vi.axes_manager.signal_shape == (2, 2)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""This contains the data and registry for the data shipped with pyXem.
This data can be used to test pyXem and to provide examples.

"""


_zenodo_url = "https://zenodo.org/records/15490547/files"  # version 0.9.0
# file name : hash
_file_names_hash = {
    "au_xgrating_100kX.hspy": "md5:b0733af9d0a272fc1a47e2f56a324fe5",
    "data_processed.zspy": "md5:bfd9f7e524a65c2988e927f6422dedeb",
    "GaAs_mp-2534_conventional_standard.cif": "md5:9abcda3883bafe29a0790994782ad136",
    "GaAs_mp-8883_conventional_standard.cif": "md5:0116734c35e2a221d08e5e49194e3f3f",
    "MgO.cif": "md5:18224100df2e2b72ad9127bc9f201228",
    "mgo_nano.zspy": "md5:38a03c23cec147a9cffab9ea349cd15a",
    "PdNiP.zspy": "md5:b16375d23eda3da60ea2ff36cc11c5bd",
    "sample_with_g.zspy": "md5:65c0a17387a10ca21ebf1812bab67568",
    "twinned_nanowire.hdf5": "md5:2765fa855db8bc011252dbb425facc81",
    "ZrNbPercipitate.zspy": "md5:7012a2bdf564f4fb25299af05412723f",
    "cuzipProcessed.zspy": "md5:829ecb8f765acb6e9a22092339c6a268",
    "colorwheel.txt": "md5:1555136e42cae858be0716f007bda4e4",
    "SPED-Ag.zspy": "md5:8556346543fc19f0ef9bdc0f4a6619b5",
    "PdCuSiCrystalization-zip.zspy": "md5:80ec7f95ec250106c586debf5d814325",
    "FeAl_stripes.zspy": "md5:702cb0c8ff75062c0cb23b3722c2f859",
    "au_xgrating_20cm.tif": "md5:06192653b9f7841f16a29d3d04e0fd06",
    "au.cif": "md5:4cb2856e8ed9ffac34f5fa22424cd2a2",
    "smallPtychography.hspy": "md5:df9376d5c020a23f0f7f51cfe79f303f",
    "pos6-1.zspy": "md5:102d95883974acb8f6698dcec52676c7",
    "pos5-1.zspy": "md5:351e8bfc12f2628edc3009cd5a429f4d",
    "pos1-1.zspy": "md5:2195347310aa35d6835dc141d07f783f",
}
# add to _urls and to _file_names_hash
# if you want to download the data from a different location
_urls = {key: _zenodo_url + "/" + key for key in _file_names_hash.keys()}



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from pyxem.utils.virtual_images_utils import get_vectors_mesh


def test_get_vectors_mesh():
    mesh1 = np.array(
        [
            [-1.0, -1.0],
            [0.0, -1.0],
            [1.0, -1.0],
            [-1.0, 0.0],
            [0.0, 0.0],
            [1.0, 0.0],
            [-1.0, 1.0],
            [0.0, 1.0],
            [1.0, 1.0],
        ]
    )
    mesh = get_vectors_mesh(1.0, 1.0, g_norm_max=1.5, angle=0.0, shear=0.0)
    np.testing.assert_allclose(mesh, mesh1)

    mesh = get_vectors_mesh(2.0, 2.0, g_norm_max=3.0, angle=0.0, shear=0.0)
    np.testing.assert_allclose(mesh, mesh1 * 2)

    mesh = get_vectors_mesh(1.5, 1.0, g_norm_max=1.9, angle=0.0, shear=0.0)
    np.testing.assert_allclose(mesh.T[0], mesh1.T[0] * 1.5)
    np.testing.assert_allclose(mesh.T[1], mesh1.T[1])

    mesh2 = np.array(
        [
            [0.5, -0.8660254],
            [-0.8660254, -0.5],
            [0.0, 0.0],
            [0.8660254, 0.5],
            [-0.5, 0.8660254],
        ]
    )

    mesh = get_vectors_mesh(1.0, 1.0, g_norm_max=1.0, angle=30, shear=0.0)
    np.testing.assert_allclose(mesh, mesh2)

    mesh3 = np.array([[-0.1, -1.0], [-1.0, 0.0], [0.0, 0.0], [1.0, 0.0], [0.1, 1.0]])

    mesh = get_vectors_mesh(1.0, 1.0, g_norm_max=1.2, angle=0.0, shear=0.1)
    np.testing.assert_allclose(mesh, mesh3)

    mesh4 = np.array(
        [
            [0.49607695, -1.09923048],
            [1.36210236, -0.59923048],
            [-0.8660254, -0.5],
            [0.0, 0.0],
            [0.8660254, 0.5],
            [-1.36210236, 0.59923048],
            [-0.49607695, 1.09923048],
        ]
    )

    mesh = get_vectors_mesh(1.0, 1.2, g_norm_max=1.5, angle=30.0, shear=0.1)
    np.testing.assert_allclose(mesh, mesh4)

    with pytest.raises(ValueError):
        get_vectors_mesh(1.0, 1.0, g_norm_max=1.5, angle=0.0, shear=2.0)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da

from pyxem.signals import Power2D, LazyPower2D


class TestComputeAndAsLazy2D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyPower2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert s.__class__ == Power2D
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_4d_data_compute(self):
        dask_array = da.random.random((4, 4, 10, 15), chunks=(1, 1, 10, 15))
        s = LazyPower2D(dask_array)
        s.compute()
        assert s.__class__ == Power2D
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = Power2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyPower2D
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_4d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = Power2D(data)
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyPower2D
        assert data.shape == s_lazy.data.shape


class TestPower:
    @pytest.fixture
    def flat_pattern(self):
        pd = Power2D(data=np.ones(shape=(2, 2, 5, 5)))
        return pd

    @pytest.mark.parametrize("k_region", [None, [2.0, 4.0]])
    @pytest.mark.parametrize("sym", [None, 4, [2, 4]])
    def test_power_signal_get_map(self, flat_pattern, k_region, sym):
        flat_pattern.get_map(k_region=k_region, symmetry=sym)

    @pytest.mark.parametrize("k_region", [None, [2.0, 4.0]])
    @pytest.mark.parametrize("sym", [[2, 4]])
    def test_power_signal_plot_symmetries(self, flat_pattern, k_region, sym):
        flat_pattern.plot_symmetries(k_region=k_region, symmetry=sym)


class TestDecomposition:
    def test_decomposition_is_performed(self, diffraction_pattern):
        s = Power2D(diffraction_pattern)
        s.decomposition()
        assert s.learning_results is not None

    def test_decomposition_class_assignment(self, diffraction_pattern):
        s = Power2D(diffraction_pattern)
        s.decomposition()
        assert isinstance(s, Power2D)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from pytest import approx, mark
import math

import numpy as np
from numpy.testing import assert_allclose
from scipy.signal import convolve2d
from skimage import morphology
from hyperspy.signals import Signal2D

from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
from pyxem.signals import Diffraction2D
import pyxem.utils.ransac_ellipse_tools as ret


class TestIsEllipseGood:
    def test_simple(self):
        params = ret._make_ellipse_model_params_focus(30, 50, 30, 20, 0)
        model = ret.EllipseModel()
        model.params = params
        is_good = ret.is_ellipse_good(
            ellipse_model=model, data=None, xf=30, yf=50, rf_lim=5
        )
        assert is_good

    @mark.parametrize("xf,out", [(40, True), (46, False), (34, False)])
    def test_xf(self, xf, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(40, 50, 10, 30, 0)
        is_good = ret.is_ellipse_good(
            ellipse_model=model, data=None, xf=xf, yf=50, rf_lim=5
        )
        assert is_good == out

    @mark.parametrize("yf,out", [(20, False), (10, True), (-9, False)])
    def test_yf(self, yf, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(70, 10, 10, 30, 0)
        is_good = ret.is_ellipse_good(
            ellipse_model=model, data=None, xf=70, yf=yf, rf_lim=5
        )
        assert is_good == out

    @mark.parametrize("rf_lim,out", [(2, False), (4, True), (3.5, True)])
    def test_rf_lim(self, rf_lim, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(47, 30, 10, 30, 0)
        is_good = ret.is_ellipse_good(
            ellipse_model=model, data=None, xf=50, yf=30, rf_lim=rf_lim
        )
        assert is_good == out

    @mark.parametrize("semi_len_min,out", [(4, True), (7, False), (6.5, True)])
    def test_semi_len_min(self, semi_len_min, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(50, 30, 6.6, 10, 0)
        is_good = ret.is_ellipse_good(
            ellipse_model=model,
            data=None,
            xf=50,
            yf=30,
            rf_lim=5,
            semi_len_min=semi_len_min,
        )
        assert is_good == out

    @mark.parametrize(
        "a,b,semi_len_max,out",
        [
            (3, 6.5, 6.6, True),
            (3, 6.5, 4, False),
            (3, 6.5, 9, True),
            (6.5, 3, 4, False),
        ],
    )
    def test_semi_len_max(self, a, b, semi_len_max, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(50, 30, a, b, 0)
        params = list(model.params)
        # Changing the params like this is necessary to test if b is larger
        # than semi_len_max. This is due to the content in model.params not being
        # the same as the input in ret._make_ellipse_model_params_focus.
        params[2:5] = a, b, 0.0
        model.params = tuple(params)
        is_good = ret.is_ellipse_good(
            ellipse_model=model,
            data=None,
            xf=50,
            yf=30,
            rf_lim=5,
            semi_len_max=semi_len_max,
        )
        assert is_good == out

    @mark.parametrize("semi_len_ratio_lim,out", [(2.1, True), (1.5, False), (3, True)])
    def test_semi_len_ratio_lim(self, semi_len_ratio_lim, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(50, 30, 9, 4.5, 0)
        is_good = ret.is_ellipse_good(
            ellipse_model=model,
            data=None,
            xf=50,
            yf=30,
            rf_lim=5,
            semi_len_ratio_lim=semi_len_ratio_lim,
        )
        assert is_good == out

    @mark.parametrize("r", [0, 0.5, 0.9, 1.3, np.pi, 2 * np.pi, 9.9])
    def test_different_rotations(self, r):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(50, 30, 3, 6.5, r)
        is_good = ret.is_ellipse_good(
            ellipse_model=model, data=None, xf=50, yf=30, rf_lim=5
        )
        assert is_good

    @mark.parametrize(
        "x, y, r, smin, smax, srl, out",
        [
            (5, 3, 1, 2, 7, 2, True),
            (3, 3, 1, 2, 7, 2, False),
            (5, 5, 1, 2, 7, 2, False),
            (5, 2, 3, 2, 7, 2, True),
            (5, 2, 3, 5, 7, 2, False),
            (5, 2, 3, 5, 6, 2, False),
            (5, 3, 1, 3, 7, 1.5, False),
            (1, 9, 1, 5, 6, 1.1, False),
        ],
    )
    def test_many_different_ones(self, x, y, r, smin, smax, srl, out):
        model = ret.EllipseModel()
        model.params = ret._make_ellipse_model_params_focus(5, 3, 4, 6.5, 0.1)
        is_good = ret.is_ellipse_good(
            ellipse_model=model,
            data=None,
            xf=x,
            yf=y,
            rf_lim=r,
            semi_len_min=smin,
            semi_len_max=smax,
            semi_len_ratio_lim=srl,
        )
        assert is_good == out


class TestMakeEllipseDataPoints:
    def test_simple(self):
        data = ret.make_ellipse_data_points(5, 2, 9, 5, 0)
        assert data.size > 0

    def test_nt(self):
        data0 = ret.make_ellipse_data_points(5, 2, 9, 5, 0, nt=10)
        assert data0.shape == (10, 2)
        data1 = ret.make_ellipse_data_points(5, 2, 9, 5, 0, nt=29)
        assert data1.shape == (29, 2)

    def test_use_focus(self):
        x, y, a, b, r = 5, 9, 10, 8, 0
        data0 = ret.make_ellipse_data_points(x, y, a, b, r, nt=99, use_focus=False)
        assert np.allclose(data0.mean(axis=0), (x, y))
        data1 = ret.make_ellipse_data_points(x, y, a, b, r, nt=99, use_focus=True)
        assert not np.allclose(data1.mean(axis=0), (x, y))

    def test_xy_use_focus_false(self):
        x0, y0, x1, y1 = 10, -5, -20, 30
        a, b, r = 10, 8, 0
        data0 = ret.make_ellipse_data_points(x0, y0, a, b, r, nt=99, use_focus=False)
        assert np.allclose(data0.mean(axis=0), (x0, y0))
        data1 = ret.make_ellipse_data_points(x1, y1, a, b, r, nt=99, use_focus=False)
        assert np.allclose(data1.mean(axis=0), (x1, y1))

    def test_xy_use_focus_true(self):
        x0, y0, x1, y1 = 10, -5, -20, 30
        a, b, r = 10, 8, 0
        data0 = ret.make_ellipse_data_points(x0, y0, a, b, r, nt=99, use_focus=True)
        xc0, yc0 = data0.mean(axis=0)
        f0 = ret._get_closest_focus(x0, y0, xc0, yc0, a, b, r)
        assert approx(f0) == (x0, y0)
        data1 = ret.make_ellipse_data_points(x1, y1, a, b, r, nt=99, use_focus=True)
        xc1, yc1 = data1.mean(axis=0)
        f1 = ret._get_closest_focus(x1, y1, xc1, yc1, a, b, r)
        assert approx(f1) == (x1, y1)

    def test_ab(self):
        x, y, r, nt = 10, 20, 0, 9999
        a0, b0, a1, b1 = 10, 5, 8, 15
        data0 = ret.make_ellipse_data_points(x, y, a0, b0, r, nt=nt, use_focus=False)
        assert np.allclose(data0.min(axis=0), (x - a0, y - b0), atol=10e-5)
        data1 = ret.make_ellipse_data_points(x, y, a1, b1, r, nt=nt, use_focus=False)
        assert np.allclose(data1.min(axis=0), (x - a1, y - b1), atol=10e-5)

    def test_r(self):
        x, y, a, b, nt = 10, 20, 15, 5, 9999
        r0, r1 = 0, math.pi / 2
        data0 = ret.make_ellipse_data_points(x, y, a, b, r0, nt=nt, use_focus=False)
        assert np.allclose(data0.min(axis=0), (x - a, y - b), atol=10e-5)
        data1 = ret.make_ellipse_data_points(x, y, a, b, r1, nt=nt, use_focus=False)
        assert np.allclose(data1.min(axis=0), (x - b, y - a), atol=10e-5)


class TestGetClosestFocus:
    def test_circle(self):
        xc, yc = 10, 20
        x, y, a, b, r = 10, 20, 10, 10, 0
        xf, yf = ret._get_closest_focus(xc, yc, x, y, a, b, r)
        assert (xf, yf) == (10, 20)

    def test_horizontal_ellipse(self):
        x, y, a, b, r = 10, 20, 20, 10, 0
        c = math.sqrt(a**2 - b**2)
        xf0, yf0 = ret._get_closest_focus(20, 20, x, y, a, b, r)
        assert (xf0, yf0) == (x + c, y)
        xf1, yf1 = ret._get_closest_focus(5, 20, x, y, a, b, r)
        assert (xf1, yf1) == (x - c, y)

    def test_vertical_ellipse(self):
        x, y, a, b, r = 10, 20, 15, 10, math.pi / 2
        c = math.sqrt(a**2 - b**2)
        xf0, yf0 = ret._get_closest_focus(10, 30, x, y, a, b, r)
        assert (xf0, yf0) == (x, y + c)
        xf1, yf1 = ret._get_closest_focus(10, 10, x, y, a, b, r)
        assert (xf1, yf1) == (x, y - c)


class TestEllipseCentreToFocus:
    def test_circle(self):
        x, y, a, b, r = 10, 20, 10, 10, 0
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        assert f0 == (x, y)
        assert f1 == (x, y)

    def test_horizontal_ellipse(self):
        x, y, a, b, r = 10, 20, 20, 10, 0
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        c = math.sqrt(a**2 - b**2)
        assert f0 == (x + c, y)
        assert f1 == (x - c, y)

    def test_vertical_ellipse(self):
        x, y, a, b, r = 10, 20, 10, 5, math.pi / 2
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        c = math.sqrt(a**2 - b**2)
        assert f0 == (x, y + c)
        assert f1 == (x, y - c)

    def test_rotated45_ellipse(self):
        x, y, a, b, r = 10, 20, 10, 5, math.pi / 4
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        c = math.sqrt(a**2 - b**2)
        assert approx(f0) == (x + c * math.sin(r), y + c * math.cos(r))
        assert approx(f1) == (x - c * math.sin(r), y - c * math.cos(r))

    def test_rotated_negative45_ellipse(self):
        x, y, a, b, r = 10, 20, 10, 5, -math.pi / 4
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        c = math.sqrt(a**2 - b**2)
        assert approx(f0) == (x - c * math.sin(r), y - c * math.cos(r))
        assert approx(f1) == (x + c * math.sin(r), y + c * math.cos(r))

    def test_horizontal_negative(self):
        x, y, a, b, r = 5, 20, 20, 10, 0
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        c = math.sqrt(a**2 - b**2)
        assert approx(f0) == (x + c, y)
        assert approx(f1) == (x - c, y)

    def test_vertical_negative(self):
        x, y, a, b, r = 10, 5, 10, 20, 0
        f0, f1 = ret._ellipse_centre_to_focus(x, y, a, b, r)
        c = math.sqrt(b**2 - a**2)
        assert approx(f0) == (x, y + c)
        assert approx(f1) == (x, y - c)


def compare_model_params(params0, params1, rel=None, abs=None):
    xf0, yf0, a0, b0, r0 = params0
    xf1, yf1, a1, b1, r1 = params1
    # only case in the params
    if a1 < b1:
        r1 += math.pi / 2
        a1, b1 = b1, a1
    if a0 < b0:
        r0 += math.pi / 2
        a0, b0 = b0, a0
    r0 = ((r0 - np.pi / 2) % np.pi) - (np.pi / 2)
    r1 = ((r1 - np.pi / 2) % np.pi) - (np.pi / 2)
    assert approx((xf0, yf0), rel=rel, abs=abs) == (xf1, yf1)
    assert approx((a0, b0), rel=rel, abs=abs) == (a1, b1)
    assert approx(r0, rel=rel, abs=abs) == r1


class TestGetEllipseModelRansacSingleFrame:
    def test_simple(self):
        data = ret.make_ellipse_data_points(50, 55, 20, 16, 2, nt=15)
        ret.get_ellipse_model_ransac_single_frame(data)

    def test_min_samples(self):
        xf, yf, a, b, r = 50, 55, 21, 20, 0
        data = ret.make_ellipse_data_points(xf, yf, a, b, r, nt=20)
        data = np.vstack((data, [70, 55]))
        model_ransac0, inliers0 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=5,
            semi_len_min=15,
            semi_len_max=None,
            semi_len_ratio_lim=None,
            min_samples=20,
            residual_threshold=3,
            max_trials=100,
        )
        assert inliers0[:-1].all()
        assert not inliers0[-1]
        model_ransac1, inliers1 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=5,
            semi_len_min=15,
            semi_len_max=None,
            semi_len_ratio_lim=None,
            min_samples=20,
            residual_threshold=6,
            max_trials=100,
        )
        assert inliers1.all()

    def test_min_samples_smaller_than_data(self):
        data = ret.make_ellipse_data_points(50, 55, 20, 16, 2, nt=15)
        ret.get_ellipse_model_ransac_single_frame(data, min_samples=25)

    @mark.skip(reason="Broken due to changes in skimage 0.19")
    def test_all_inliers(self):
        xf, yf, a, b, r = 50, 55, 20, 16, 2
        data = ret.make_ellipse_data_points(xf, yf, a, b, r, nt=25)
        model_ransac, inliers = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=2,
            semi_len_min=10,
            semi_len_max=25,
            semi_len_ratio_lim=2,
            min_samples=6,
            residual_threshold=10,
            max_trials=100,
        )
        params0f = (xf, yf, a, b, r)
        params1f = ret._ellipse_model_centre_to_focus(*model_ransac.params, xf, yf)
        assert inliers.all()
        compare_model_params(params0f, params1f, abs=0.01)

    @mark.skip(reason="Broken due to changes in skimage 0.19")
    def test_xf(self):
        xf0, xf1, yf, a, b, r = 50, 120, 55, 21, 20, 2
        data0 = ret.make_ellipse_data_points(xf0, yf, a, b, r, nt=25)
        data1 = ret.make_ellipse_data_points(xf1, yf, a, b, r, nt=25)
        data = np.vstack((data0, data1))
        model_ransac0, inliers0 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=xf0,
            yf=55,
            rf_lim=2,
            semi_len_min=19,
            semi_len_max=None,
            semi_len_ratio_lim=2,
            min_samples=5,
            residual_threshold=10,
            max_trials=300,
        )
        params00f = (xf0, yf, a, b, r)
        params01f = ret._ellipse_model_centre_to_focus(*model_ransac0.params, xf0, yf)
        compare_model_params(params00f, params01f, abs=0.1)

        model_ransac1, inliers1 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=xf1,
            yf=55,
            rf_lim=2,
            semi_len_min=19,
            semi_len_max=None,
            semi_len_ratio_lim=2,
            min_samples=5,
            residual_threshold=10,
            max_trials=300,
        )
        params10f = (xf1, yf, a, b, r)
        params11f = ret._ellipse_model_centre_to_focus(*model_ransac1.params, xf1, yf)
        compare_model_params(params10f, params11f, abs=0.1)

    @mark.skip(reason="Broken due to changes in skimage 0.19")
    def test_yf(self):
        xf, yf0, yf1, a, b, r = 50, 55, 140, 21, 20, 2
        data0 = ret.make_ellipse_data_points(xf, yf0, a, b, r, nt=25)
        data1 = ret.make_ellipse_data_points(xf, yf1, a, b, r, nt=25)
        data = np.vstack((data0, data1))
        model_ransac0, inliers0 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=yf0,
            rf_lim=2,
            semi_len_min=19,
            semi_len_max=None,
            semi_len_ratio_lim=2,
            min_samples=5,
            residual_threshold=10,
            max_trials=300,
        )
        params00f = (xf, yf0, a, b, r)
        params01f = ret._ellipse_model_centre_to_focus(*model_ransac0.params, xf, yf0)
        compare_model_params(params00f, params01f, abs=0.1)

        model_ransac1, inliers1 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=yf1,
            rf_lim=2,
            semi_len_min=19,
            semi_len_max=None,
            semi_len_ratio_lim=2,
            min_samples=5,
            residual_threshold=10,
            max_trials=300,
        )
        params10f = (xf, yf1, a, b, r)
        params11f = ret._ellipse_model_centre_to_focus(*model_ransac1.params, xf, yf1)
        compare_model_params(params10f, params11f, abs=0.1)

    def test_rf_lim(self):
        xf, yf, a, b, r = 50, 55, 21, 20, 2
        data = ret.make_ellipse_data_points(xf, yf, a, b, r, nt=25)
        model_ransac0, inliers0 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=50,
            rf_lim=6,
            semi_len_min=15,
            semi_len_max=23,
            semi_len_ratio_lim=1.2,
            min_samples=5,
            residual_threshold=10,
            max_trials=100,
        )
        params0f = (xf, yf, a, b, r)
        params1f = ret._ellipse_model_centre_to_focus(*model_ransac0.params, xf, yf)
        assert inliers0.all()
        compare_model_params(params0f, params1f, abs=0.1)

        model_ransac1, inliers1 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=50,
            rf_lim=3,
            semi_len_min=15,
            semi_len_max=23,
            semi_len_ratio_lim=1.2,
            min_samples=5,
            residual_threshold=10,
            max_trials=100,
        )
        assert model_ransac1 is None

    def test_semi_len_min(self):
        xf, yf, a, b, r = 50, 55, 21, 20, 2
        data = ret.make_ellipse_data_points(xf, yf, a, b, r, nt=25)
        model_ransac0, inliers0 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=2,
            semi_len_min=19,
            semi_len_max=None,
            semi_len_ratio_lim=2,
            min_samples=6,
            residual_threshold=10,
            max_trials=100,
        )
        params0f = (xf, yf, a, b, r)
        params1f = ret._ellipse_model_centre_to_focus(*model_ransac0.params, xf, yf)
        assert inliers0.all()
        compare_model_params(params0f, params1f, abs=0.1)

        model_ransac1, inliers1 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=2,
            semi_len_min=25,
            semi_len_max=None,
            semi_len_ratio_lim=2,
            min_samples=6,
            residual_threshold=10,
            max_trials=100,
        )
        assert model_ransac1 is None

    def test_semi_len_max(self):
        xf, yf, a, b, r = 50, 55, 21, 20, 2
        data = ret.make_ellipse_data_points(xf, yf, a, b, r, nt=25)
        model_ransac0, inliers0 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=2,
            semi_len_min=None,
            semi_len_max=25,
            semi_len_ratio_lim=2,
            min_samples=6,
            residual_threshold=10,
            max_trials=100,
        )
        params0f = (xf, yf, a, b, r)
        params1f = ret._ellipse_model_centre_to_focus(*model_ransac0.params, xf, yf)
        assert inliers0.all()
        compare_model_params(params0f, params1f, abs=0.1)

        model_ransac1, inliers1 = ret.get_ellipse_model_ransac_single_frame(
            data,
            xf=50,
            yf=55,
            rf_lim=2,
            semi_len_min=None,
            semi_len_max=15,
            semi_len_ratio_lim=2,
            min_samples=6,
            residual_threshold=10,
            max_trials=100,
        )
        assert model_ransac1 is None


class TestGetEllipseModelRansac:
    def test_simple(self):
        xc, yc = np.ones((2, 3)), np.ones((2, 3))
        semi0, semi1 = np.ones((2, 3)), np.ones((2, 3))
        rot = np.zeros((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(xc, yc, semi0, semi1, rot)
        ellipse_array, inlier_array = ret.get_ellipse_model_ransac(
            peak_array, max_trials=50
        )
        assert ellipse_array.shape == xc.shape
        assert inlier_array.shape == xc.shape

    def test_xc_yc(self):
        np.random.seed(7)
        xf = np.random.randint(90, 100, size=(2, 3))
        yf = np.random.randint(110, 120, size=(2, 3))
        semi0, semi1 = np.ones((2, 3)) * 60, np.ones((2, 3)) * 60
        rot = np.zeros((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi0, semi1, rot, nt=20
        )
        ellipse_array0, inlier_array0 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=95,
            yf=115,
            rf_lim=20,
            semi_len_min=50,
            semi_len_max=65,
            semi_len_ratio_lim=1.2,
            min_samples=15,
            max_trials=20,
        )
        ellipse_array1, inlier_array1 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=10,
            yf=12,
            rf_lim=20,
            semi_len_min=50,
            semi_len_max=65,
            semi_len_ratio_lim=1.2,
            min_samples=15,
            max_trials=20,
        )

        for iy, ix in np.ndindex(xf.shape):
            assert np.isclose(
                xf[iy, ix], ellipse_array0[iy, ix][0], atol=1e-8, rtol=1e-3
            )
            assert np.isclose(
                yf[iy, ix], ellipse_array0[iy, ix][1], atol=1e-8, rtol=1e-3
            )
            assert inlier_array0[iy, ix].all()
            assert ellipse_array1[iy, ix] is None
            assert inlier_array1[iy, ix] is None

    def test_semi_lengths(self):
        xf, yf = np.ones((2, 3)) * 200, np.ones((2, 3)) * 210
        np.random.seed(7)
        semi0 = np.random.randint(90, 110, size=(2, 3))
        semi1 = np.random.randint(130, 140, size=(2, 3))
        rot = np.zeros((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi0, semi1, rot, nt=20
        )
        ellipse_array, inlier_array = ret.get_ellipse_model_ransac(
            peak_array,
            xf=200,
            yf=210,
            rf_lim=20,
            semi_len_min=80,
            semi_len_max=150,
            semi_len_ratio_lim=1.7,
            min_samples=15,
            max_trials=20,
        )
        for iy, ix in np.ndindex(xf.shape):
            semi_min = min(ellipse_array[iy, ix][2], ellipse_array[iy, ix][3])
            semi_max = max(ellipse_array[iy, ix][2], ellipse_array[iy, ix][3])
            assert approx(semi_min, abs=0.01) == semi0[iy, ix]
            assert approx(semi_max, abs=0.01) == semi1[iy, ix]

    def test_rf_lim(self):
        xf, yf = np.ones((2, 3)) * 200, np.ones((2, 3)) * 210
        np.random.seed(7)
        semi0 = np.random.randint(90, 110, size=(2, 3))
        semi1 = np.random.randint(130, 140, size=(2, 3))
        rot = np.zeros((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi0, semi1, rot, nt=20
        )
        ellipse_array0, inlier_array0 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=200,
            yf=210,
            rf_lim=20,
            semi_len_min=80,
            semi_len_max=150,
            semi_len_ratio_lim=1.7,
            min_samples=15,
            max_trials=20,
        )
        ellipse_array1, inlier_array1 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=200,
            yf=210,
            rf_lim=20,
            semi_len_min=80,
            semi_len_max=150,
            semi_len_ratio_lim=1.01,
            min_samples=15,
            max_trials=20,
        )
        for iy, ix in np.ndindex(xf.shape):
            ellipse_params0 = ellipse_array0[iy, ix]
            ellipse_params1 = ellipse_array1[iy, ix]
            assert ellipse_params0 != ellipse_params1

    def test_semi_len_ratio_lim(self):
        xf, yf = np.ones((2, 3)) * 200, np.ones((2, 3)) * 210
        semi00, semi01 = np.ones((2, 3)) * 100, np.ones((2, 3)) * 100
        semi10, semi11 = np.ones((2, 3)) * 100, np.ones((2, 3)) * 190
        rot = np.zeros((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi00, semi01, rot, nt=20
        )
        peak_array1 = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi10, semi11, rot, nt=20
        )
        for iy, ix in np.ndindex(xf.shape):
            peak_array[iy, ix] = np.vstack((peak_array[iy, ix], peak_array1[iy, ix]))
        ellipse_array0, inlier_array0 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=200,
            yf=210,
            rf_lim=20,
            semi_len_min=95,
            semi_len_max=195,
            semi_len_ratio_lim=1.11,
            min_samples=15,
            max_trials=200,
        )
        ellipse_array1, inlier_array1 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=200,
            yf=210,
            rf_lim=20,
            semi_len_min=95,
            semi_len_max=195,
            semi_len_ratio_lim=2.0,
            min_samples=15,
            max_trials=200,
        )
        semi_len_ratio_list = []
        for iy, ix in np.ndindex(xf.shape):
            if ellipse_array1[iy, ix] is not None:
                semi0, semi1 = ellipse_array1[iy, ix][2:4]
                semi_len_ratio = max(semi0, semi1) / min(semi0, semi1)
                semi_len_ratio_list.append(semi_len_ratio)
        semi_len_ratio_list = np.array(semi_len_ratio_list)
        assert (semi_len_ratio_list > 1.8).any()

    def test_residual_threshold(self):
        xyf, semi = np.ones((2, 3)) * 100, np.ones((2, 3)) * 90
        rot = np.zeros((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(
            xyf, xyf, semi, semi, rot, nt=20
        )
        for iy, ix in np.ndindex(xyf.shape):
            peak_array[iy, ix] = np.vstack((peak_array[iy, ix], [100, 5]))
        ellipse_array0, inlier_array0 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=100,
            yf=100,
            semi_len_min=85,
            semi_len_max=95,
            semi_len_ratio_lim=1.1,
            residual_threshold=1,
            max_trials=100,
            min_samples=15,
        )
        ellipse_array1, inlier_array1 = ret.get_ellipse_model_ransac(
            peak_array,
            xf=100,
            yf=100,
            semi_len_min=85,
            semi_len_max=95,
            semi_len_ratio_lim=1.1,
            residual_threshold=5,
            max_trials=100,
            min_samples=15,
        )

        for iy, ix in np.ndindex(xyf.shape):
            inlier0 = inlier_array0[iy, ix]
            inlier1 = inlier_array1[iy, ix]
            assert inlier0.sum() == 20
            assert inlier1.sum() == 21

    def test_no_values_out_of_bounds(self):
        xf, yf, rf_lim = 100, 100, 20
        semi_len_min, semi_len_max = 50, 100
        semi_len_ratio_lim = 1.15
        peak_array = np.empty((10, 11), dtype=object)
        for ind in np.ndindex(peak_array.shape):
            peak_array[ind] = np.random.randint(0, 200, size=(200, 2))
        ellipse_array, inlier_array = ret.get_ellipse_model_ransac(
            peak_array,
            xf=xf,
            yf=yf,
            rf_lim=rf_lim,
            semi_len_min=semi_len_min,
            semi_len_max=semi_len_max,
            semi_len_ratio_lim=semi_len_ratio_lim,
            min_samples=15,
            max_trials=5,
        )
        for iy, ix in np.ndindex(peak_array.shape[:2]):
            if ellipse_array[iy, ix] is not None:
                xc, yc, semi0, semi1, r = ellipse_array[iy, ix]
                x, y = ret._get_closest_focus(xf, yf, xc, yc, semi0, semi1, r)
                rf = math.hypot(x - xf, y - yf)
                semi_ratio = max(semi0, semi1) / min(semi0, semi1)
                assert rf < rf_lim
                assert semi_ratio < semi_len_ratio_lim
                assert semi0 > semi_len_min
                assert semi1 > semi_len_min
                assert semi0 < semi_len_max
                assert semi1 < semi_len_max


def test_full_ellipse_ransac_processing():
    xf, yf, a, b, r, nt = 100, 115, 45, 35, 0, 15
    data_points = ret.make_ellipse_data_points(xf, yf, a, b, r, nt)
    image = np.zeros(shape=(200, 210), dtype=np.float32)
    for x, y in data_points:
        image[int(round(x)), int(round(y))] = 100
    disk = morphology.disk(5, np.uint16)
    image = convolve2d(image, disk, mode="same")
    data = np.zeros((2, 3, 210, 200), dtype=np.float32)
    data[:, :] = image.T

    s = Diffraction2D(data)
    s_t = s.template_match_disk(disk_r=5)
    peak_array = s_t.find_peaks(
        method="difference_of_gaussian",
        threshold=0.1,
        lazy_output=False,
        interactive=False,
    ).data

    c = math.sqrt(math.pow(a, 2) - math.pow(b, 2))
    xc, yc = xf - c * math.cos(r), yf - c * math.sin(r)

    for iy, ix in np.ndindex(peak_array.shape):
        peaks = peak_array[iy, ix]
        assert len(peaks) == 15
        assert approx(peaks[:, 1].mean(), abs=2) == xc
        assert approx(peaks[:, 0].mean(), abs=2) == yc
        assert approx(peaks[:, 1].max(), abs=2) == xc + a
        assert approx(peaks[:, 0].max(), abs=2) == yc + b
        assert approx(peaks[:, 1].min(), abs=2) == xc - a
        assert approx(peaks[:, 0].min(), abs=2) == yc - b

    ellipse_array, inlier_array = ret.get_ellipse_model_ransac(
        peak_array,
        yf=yf,
        xf=xf,
        rf_lim=15,
        semi_len_min=min(a, b) - 5,
        semi_len_max=max(a, b) + 5,
        semi_len_ratio_lim=5,
        max_trials=50,
        min_samples=10,
    )

    s.add_ellipse_array_as_markers(ellipse_array)

    for iy, ix in np.ndindex(ellipse_array.shape):
        ycf, xcf, bf, af, rf = ellipse_array[iy, ix]
        # shifted by pi/2
        compare_model_params(
            (xcf, ycf, af, bf, rf + np.pi / 2), (yc, xc, a, b, r), abs=0.1
        )
        assert inlier_array[iy, ix].all()

    s.add_ellipse_array_as_markers(ellipse_array)
    x_list, y_list = [], []


class TestDetermineEllipse:
    @mark.parametrize("mask", [True, False])
    @mark.parametrize("return_params", [True, False])
    @mark.parametrize("guess_starting_params", [True, False])
    def test_determine_ellipse_ransac(self, mask, return_params, guess_starting_params):
        if mask:
            mask = np.zeros((100, 100), dtype=bool)
        else:
            mask = None
        print(mask, return_params, guess_starting_params)
        t = np.ones((100, 100))
        x, y = np.ogrid[-45:55, -50:50]
        t[x**2 + (y * 1.15) ** 2 < 40**2] = 100
        t[x**2 + (y * 1.15) ** 2 < 30**2] = 1
        rng = np.random.RandomState(3)
        t = t + rng.random((100, 100))
        if not guess_starting_params:
            params = {
                "xf": 50,
                "yf": 50,
                "rf_lim": 20,
                "semi_len_min": 30,
                "semi_len_max": 100,
                "semi_len_ratio_lim": 1.2,
                "min_samples": 6,
                "residual_threshold": 20,
                "max_trials": 1000,
            }
        else:
            params = {}
        ans = ret.determine_ellipse(
            t,
            mask=mask,
            use_ransac=True,
            num_points=50,
            return_params=return_params,
            guess_starting_params=guess_starting_params,
            **params
        )
        center = ans[0]
        affine = ans[1]

        np.testing.assert_array_almost_equal(
            affine,
            [[1.0, 0, 0], [0, 0.87, 0], [0, 0, 1]],
            1,
        )
        np.testing.assert_array_almost_equal(center, [45, 50], 0)

    @mark.parametrize("mask", [True, False])
    @mark.parametrize("return_params", [True, False])
    def test_determine_ellipse_no_ransac(self, mask, return_params):
        if mask:
            mask = np.zeros((100, 100), dtype=bool)
            mask[40:50, :] = True
        else:
            mask = None
        t = np.ones((100, 100))
        x, y = np.ogrid[-45:55, -50:50]
        t[x**2 + (y * 1.15) ** 2 < 40**2] = 100
        t[x**2 + (y * 1.15) ** 2 < 30**2] = 1
        t = t + np.random.random((100, 100))
        ans = ret.determine_ellipse(
            t,
            mask=mask,
            use_ransac=False,
            return_params=return_params,
        )
        center = ans[0]
        affine = ans[1]

        np.testing.assert_array_almost_equal(
            affine,
            [[1.0, 0, 0], [0, 0.87, 0], [0, 0, 1]],
            2,
        )
        np.testing.assert_array_almost_equal(center, [45, 50], 0)

    @mark.parametrize("execution_number", range(5))
    def test_determine_ellipse_rotated(self, execution_number):
        angle = np.random.random() * np.pi
        test_data = mdtd.MakeTestData(200, 200, default=False)
        test_data.add_disk(x0=100, y0=100, r=5, intensity=30)
        test_data.add_ring_ellipse(
            x0=100, y0=100, semi_len0=64, semi_len1=70, rotation=angle
        )
        s = test_data.signal
        s.set_signal_type("electron_diffraction")

        mask = np.zeros_like(s.data, dtype=bool)
        mask[100 - 20 : 100 + 20, 100 - 20 : 100 + 20] = True
        center, affine = ret.determine_ellipse(
            s, mask=mask, use_ransac=False, num_points=2000
        )
        s.calibration(
            center=center, affine=affine, scale=0.1, units="k_nm^-1", beam_energy=200
        )
        s_az = s.get_azimuthal_integral2d(npt=100)
        assert np.sum((s_az.sum(axis=0).isig[6:] > 1).data) < 12

    @mark.parametrize("rot", np.linspace(0, 2 * np.pi, 10))
    def test_determine_ellipse_ring(self, rot):
        test_data = mdtd.MakeTestData(200, 200, default=False)
        test_data.add_disk(x0=100, y0=100, r=5, intensity=30)
        test_data.add_ring_ellipse(
            x0=100, y0=100, semi_len0=63, semi_len1=70, rotation=rot
        )
        s = test_data.signal
        s.set_signal_type("electron_diffraction")
        mask = np.zeros_like(s.data, dtype=bool)
        mask[100 - 20 : 100 + 20, 100 - 20 : 100 + 20] = True
        center, affine = ret.determine_ellipse(s, mask=mask, use_ransac=False)
        s.calibration(
            center=center, affine=affine, scale=0.1, units="k_nm^-1", beam_energy=200
        )
        s_az = s.get_azimuthal_integral2d(npt=100)
        assert np.sum((s_az.sum(axis=0).isig[10:] > 1).data) < 11

    def test_get_max_pos(self):
        t = np.ones((100, 100))
        x, y = np.ogrid[-45:55, -50:50]
        t[x**2 + (y * 1.15) ** 2 < 40**2] = 100
        t[x**2 + (y * 1.15) ** 2 < 30**2] = 1
        t = t + np.random.random((100, 100))
        s = Signal2D(t)
        pos = ret._get_max_positions(s, num_points=100)
        dist = np.array([((p[0] - 45) ** 2 + (p[1] - 50) ** 2) ** 0.5 for p in pos])
        np.testing.assert_array_less(dist, 40)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from scipy.ndimage import gaussian_filter

from hyperspy.signals import BaseSignal

from pyxem.generators import IntegrationGenerator
from pyxem.signals import DiffractionVectors, ElectronDiffraction2D


@pytest.mark.parametrize("radius, offset", [[1, 0], [2, 1], [3, 2]])
def test_integration_generator(radius, offset):
    pixel_positions = np.empty((2,), dtype=object)
    pixel_positions[0] = np.array([[0, 0], [15, -15], [-15, 15]])
    pixel_positions[1] = np.array([[0, 0], [15, -15], [-15, 15]])
    pattern = np.zeros((2, 50, 50))
    center = np.array(pattern[0].shape) / 2
    i, j = (pixel_positions[0] + center + offset).T.astype(int)
    pattern[0, i, j] = 1
    pattern[1, i, j] = 1

    dv = DiffractionVectors(pixel_positions, ragged=True)
    dp = ElectronDiffraction2D(pattern)
    ig = IntegrationGenerator(dp, dv)
    assert isinstance(ig, IntegrationGenerator)

    inties = ig.extract_intensities(radius=radius)
    assert isinstance(inties, BaseSignal)

    assert np.allclose(inties.data, [1, 1, 1])


@pytest.mark.skip(reason="Broken due to changes in HyperSpy")
def test_integration_generator_summation_method():
    pixel_positions = np.empty((2,), dtype=object)
    pixel_positions[0] = np.array([[0, 0], [25, -25], [-25, 25]])
    pixel_positions[1] = np.array([[0, 0], [25, -25], [-25, 25]])
    pattern = np.zeros((2, 100, 100))
    center = np.array(pattern[0].shape) / 2
    i, j = (pixel_positions[0] + center).T.astype(int)
    pattern[0, i, j] = 1.0
    pattern[1, i, j] = 1.0
    pattern = gaussian_filter(pattern, 2)

    dv = DiffractionVectors(pixel_positions, ragged=True)
    dp = ElectronDiffraction2D(pattern)
    ig = IntegrationGenerator(dp, dv)

    assert isinstance(ig, IntegrationGenerator)

    vectors = ig.extract_intensities_summation_method()

    assert np.allclose(pixel_positions[0], vectors.data[0], atol=0.05)
    assert np.allclose(vectors.data[0], pixel_positions[0], atol=0.05)
    assert np.allclose(vectors.intensities.data[0], 1.0, atol=0.05)
    assert np.allclose(vectors.sigma.data[0], 0.0, atol=0.05)
    assert isinstance(vectors, DiffractionVectors)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import dask.array as da

from hyperspy.api import interactive
from hyperspy.misc.utils import isiterable
import hyperspy.api as hs

from traits.trait_base import Undefined
from pyxem import CUPY_INSTALLED
from hyperspy.misc.utils import _get_block_pattern, add_scalar_axis
from tlz import concat

if CUPY_INSTALLED:
    import cupy as cp
from pyxem.utils.virtual_images_utils import normalize_virtual_images


OUT_SIGNAL_AXES_DOCSTRING = """out_signal_axes : None, iterable of int or string
            Specify which navigation axes to use as signal axes in the virtual
            image. If None, the two first navigation axis are used.
        """


class CommonDiffraction:
    """Common functions for all Diffraction Signals classes"""

    def to_device(self):  # pragma: no cover
        """Transfer the data to the GPU.

        Notes
        -----
        This code is currently in an experimental stage. Please report and issues you have on the pyxem github
        """
        if not CUPY_INSTALLED:
            raise ImportError(
                "The cupy package is required to use this method. "
                "Please install it using `conda install cupy`."
            )
        if not self._gpu:
            if self._lazy:
                self.data = self.data.map_blocks(
                    cp.asarray
                )  # pass chunk-wise the data to GPU
            else:
                self.data = cp.asarray(self.data)  # pass all the data to the GPU

    def to_host(self):  # pragma: no cover
        """Transfer the data from the GPU to the CPU.

        Notes
        -----
        This code is currently in an experimental stage. Please report and issues you have on the pyxem github
        """
        if not CUPY_INSTALLED:
            raise ImportError(
                "The cupy package is required to use this method. "
                "Please install it using `conda install cupy`."
            )
        if self._gpu:
            if self._lazy:
                self.data = self.data.map_blocks(cp.asnumpy)
            else:
                self.data = cp.asnumpy(self.data)

    @property
    def _gpu(self):
        if not CUPY_INSTALLED:
            return False
        else:
            if self._lazy:
                return isinstance(self.data._meta, cp.ndarray)
            else:
                return isinstance(self.data, cp.ndarray)

    @property
    def unit(self):
        if self.axes_manager.signal_axes[0].units is Undefined:
            print("The unit hasn't been set yet")
            return
        else:
            return self.axes_manager.signal_axes[0].units

    @unit.setter
    def unit(self, unit):
        """Set the units

        Parameters
        ----------
        unit : "q_nm^-1", "q_A^-1","k_nm^-1","k_A^-1","2th_deg", "2th_rad"
            The diffraction units
        """
        acceptable = ["q_nm^-1", "q_A^-1", "k_nm^-1", "k_A^-1", "2th_deg", "2th_rad"]
        if unit in acceptable:
            for axes in self.axes_manager.signal_axes:
                axes.units = unit
        else:
            print(
                'The unit must be "q_nm^-1", "q_A^-1","k_nm^-1",'
                '"k_A^-1","2th_deg", "2th_rad"'
            )

    @staticmethod
    def _get_sum_signal(signal, out_signal_axes=None):
        out = signal.nansum(signal.axes_manager.signal_axes)
        if out_signal_axes is None:
            out_signal_axes = list(
                np.arange(min(signal.axes_manager.navigation_dimension, 2))
            )
        if len(out_signal_axes) > signal.axes_manager.navigation_dimension:
            raise ValueError(
                "The length of 'out_signal_axes' can't be longer"
                "than the navigation dimension of the signal."
            )
        # Reset signal to default Signal1D or Signal2D
        out.set_signal_type("")
        return out.transpose(out_signal_axes)

    def plot_integrated_intensity(self, roi, out_signal_axes=None, **kwargs):
        """Interactively plots the integrated intensity over the scattering
        range defined by the roi.

        Parameters
        ----------
        roi : float
            Any interactive ROI detailed in HyperSpy.
        out_signal_axes : None, iterable of int or string
            Specify which navigation axes to use as signal axes in the virtual
            image. If None, the two first navigation axis are used.
        **kwargs:
            Keyword arguments to be passed to the `plot` method of the virtual
            image.

        Examples
        --------
        .. code-block:: python

            >>> # For 1D diffraction signal, we can use a SpanROI
            >>> roi = hs.roi.SpanROI(left=1., right=2.)
            >>> dp.plot_integrated_intensity(roi)

        .. code-block:: python

            >>> # For 2D diffraction signal,we can use a CircleROI
            >>> roi = hs.roi.CircleROI(3, 3, 5)
            >>> dp.plot_integrated_intensity(roi)

        """
        # Plot signal when necessary
        if self._plot is None or not self._plot.is_active:
            self.plot()

        # Get the sliced signal from the roi
        sliced_signal = roi.interactive(self, axes=self.axes_manager.signal_axes)

        # Create an output signal for the virtual dark-field calculation.
        out = self._get_sum_signal(self, out_signal_axes)
        out.metadata.General.title = "Integrated intensity"

        # Create the interactive signal
        interactive(
            sliced_signal.nansum,
            axis=sliced_signal.axes_manager.signal_axes,
            event=roi.events.changed,
            recompute_out_event=None,
            out=out,
        )

        # Plot the result
        out.plot(**kwargs)

    def get_virtual_image(self, rois, new_axis_dict=None, normalize=False):
        """Get a virtual images from a set of rois

        Parameters
        ----------
        rois : iterable of :obj:`hyperspy.roi.BaseInteractiveROI`
            Any interactive ROI detailed in HyperSpy.
        new_axis_dict : dict, optional
            A dictionary with the properties of the new axis. If None, a default
            axis is created.
        normalize : bool, optional
            If True, the virtual images are normalized to the maximum value.
        """
        if not isiterable(rois):
            rois = [
                rois,
            ]
        if new_axis_dict is None:
            new_axis_dict = {
                "name": "Virtual Dark Field",
                "offset": 0,
                "scale": 1,
                "units": "a.u.",
                "size": len(rois),
            }

        vdfs = [self.get_integrated_intensity(roi) for roi in rois]

        vdfim = hs.stack(
            vdfs, new_axis_name=new_axis_dict["name"], show_progressbar=False
        )

        vdfim.set_signal_type("virtual_dark_field")

        if vdfim.metadata.has_item("Diffraction.integrated_range"):
            del vdfim.metadata.Diffraction.integrated_range
        vdfim.metadata.set_item("Diffraction.roi_list", [f"{roi}" for roi in rois])

        # Set new axis properties
        if len(rois) > 1:
            new_axis = vdfim.axes_manager[new_axis_dict["name"]]
            for k, v in new_axis_dict.items():
                setattr(new_axis, k, v)

        if normalize:
            vdfim.map(normalize_virtual_images, show_progressbar=False)
        return vdfim

    def get_integrated_intensity(self, roi, out_signal_axes=None):
        """Obtains the intensity integrated over the scattering range as
        defined by the roi.

        Parameters
        ----------
        roi : :obj:`hyperspy.roi.BaseInteractiveROI`
            Any interactive ROI detailed in HyperSpy.
        %s

        Returns
        -------
        integrated_intensity : :obj:`hyperspy.signals.Signal2D` or :obj:`hyperspy.signals.Signal1D`
            The intensity integrated over the scattering range as defined by
            the roi.

        Examples
        --------
        .. code-block:: python

            >>> # For 1D diffraction signal, we can use a SpanROI
            >>> roi = hs.roi.SpanROI(left=1., right=2.)
            >>> virtual_image = dp.get_integrated_intensity(roi)

        .. code-block:: python

            >>> # For 2D diffraction signal,we can use a CircleROI
            >>> roi = hs.roi.CircleROI(3, 3, 5)
            >>> virtual_image = dp.get_integrated_intensity(roi)

        """
        dark_field = roi(self, axes=self.axes_manager.signal_axes)
        dark_field_sum = self._get_sum_signal(dark_field, out_signal_axes)
        dark_field_sum.metadata.General.title = "Integrated intensity"
        roi_info = f"{roi}"
        if self.metadata.get_item("General.title") not in ("", None):
            roi_info += f" of {self.metadata.General.title}"
        dark_field_sum.metadata.set_item("Diffraction.integrated_range", roi_info)

        return dark_field_sum

    get_integrated_intensity.__doc__ %= OUT_SIGNAL_AXES_DOCSTRING

    def add_navigation_signal(self, data, name="nav1", unit=None, nav_plot=False):
        """Adds in a navigation signal to the metadata.  Any type of navigation signal is acceptable.

        Parameters
        -------------------
        data: np.array
            The data for the navigation signal.  Should be the same size as the navigation axis.
        name: str
            The name of the axis.
        unit: str
            The units for the intensity of the plot. e.g 'nm' for thickness.
        """
        dict_signal = {}
        dict_signal[name] = {
            "data": data,
            "unit": unit,
            "use_as_navigation_plot": nav_plot,
        }
        if not self.metadata.has_item("Navigation_signals"):
            self.metadata.add_node("Navigation_signals")
        self.metadata.Navigation_signals.add_dictionary(dict_signal)

    def _map_blocks_prepare(self, navigation_chunks="auto"):
        """Prepare the signal for mapping blocks.  This function will check if the signal
        is chunked correctly and rechunk if necessary.
        """
        if not self._lazy:
            s_input = self.as_lazy()
            s_input.rechunk(nav_chunks=navigation_chunks)
        else:
            s_input = self
        chunk_span = np.equal(s_input.data.chunksize, s_input.data.shape)
        chunk_span = [
            chunk_span[i] for i in s_input.axes_manager.signal_indices_in_array
        ]

        if not all(chunk_span):
            old_sig = s_input.rechunk(inplace=False, nav_chunks=None)
        else:
            old_sig = s_input
        return old_sig

    def _blockwise(
        self,
        func,
        signal_shape=None,
        navigation_shape=None,
        dtype=None,
        inplace=False,
        lazy_output=None,
        ragged=False,
        num_workers=None,
        meta=None,
        **kwargs,
    ):
        """Apply a function to each block of the signal. This function might change in the
        future.

        Parameters
        ----------
        func : callable
            The function to apply to each block.
        *args : list
            The arguments to pass to the function.
        signal_shape : tuple
            The shape of the signal.
        dtype: np.dtype
            The type of the signal.
        **kwargs : dict
            The keyword arguments to pass to the function.

        Returns
        -------
        Diffraction2D
            The signal with the function applied to each block.
        """
        if dtype is None:
            dtype = self.data.dtype
        if signal_shape is None:
            signal_shape = self.axes_manager._signal_shape_in_array
        if navigation_shape is None:
            navigation_shape = self.axes_manager._navigation_shape_in_array
        if lazy_output is None:
            lazy_output = self._lazy
        new_shape = navigation_shape + signal_shape
        old_sig = self._map_blocks_prepare()
        if meta is None:
            meta = old_sig.data._meta

        arg_pairs, adjust_chunks, new_axis, output_pattern = _get_block_pattern(
            (old_sig.data,), new_shape
        )
        axes_changed = len(new_axis) != 0 or len(adjust_chunks) != 0

        mapped = da.blockwise(
            func,
            output_pattern,
            *concat(arg_pairs),
            adjust_chunks=adjust_chunks,
            new_axes=new_axis,
            align_arrays=False,
            dtype=dtype,
            concatenate=True,
            meta=meta,
            **kwargs,
        )
        data_stored = False
        if inplace:
            if (
                not self._lazy
                and not lazy_output
                and (mapped.shape == self.data.shape)
                and (mapped.dtype == self.data.dtype)
            ):
                # da.store is used to avoid unnecessary amount of memory usage.
                # By using it here, the contents in mapped is written directly to
                # the existing NumPy array, avoiding a potential doubling of memory use.
                da.store(
                    mapped,
                    self.data,
                    dtype=mapped.dtype,
                    compute=True,
                    num_workers=num_workers,
                )
                data_stored = True
            else:
                self.data = mapped
            sig = self
        else:
            sig = self._deepcopy_with_new_data(mapped)

        am = sig.axes_manager
        sig._lazy = lazy_output

        if ragged:
            axes_dicts = self.axes_manager._get_navigation_axes_dicts()
            sig.axes_manager.__init__(axes_dicts)
            sig.axes_manager._ragged = True
        elif axes_changed:
            am.remove(am.signal_axes[len(signal_shape) :])
            for ind in range(len(signal_shape) - am.signal_dimension, 0, -1):
                am._append_axis(size=signal_shape[-ind], navigate=False)

        if not ragged:
            sig.axes_manager._ragged = False
            if signal_shape == () and am.navigation_dimension == 0:
                add_scalar_axis(sig)
            sig.get_dimensions_from_data()
        sig._assign_subclass()

        if not lazy_output and not data_stored:
            sig.data = sig.data.compute(num_workers=num_workers)
        return sig



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from pyxem.data.dummy_data import CrystalSTEMSimulation

from pyxem.data import (
    simulated_overlap,
    simulated_pn_junction,
    simulated_constant_shift_magnitude,
)
from pyxem.signals import ElectronDiffraction2D, Diffraction2D


class TestMakeOverlapData:
    def test_init(self):
        s = simulated_overlap()


class TestDPC:
    def test_pn_junction(self):
        s = simulated_pn_junction()
        assert isinstance(s, ElectronDiffraction2D)
        assert s.metadata["title"] == "Simulated pn-junction"
        assert s.axes_manager.signal_axes[0].units == "nm^-1"
        assert s.axes_manager.signal_axes[1].units == "nm^-1"

    def test_simulated_constant_shift_magnitude(self):
        s = simulated_constant_shift_magnitude()
        assert isinstance(s, Diffraction2D)
        assert s.metadata["title"] == "Simulated Constant Shift Magnitude"
        assert s.axes_manager.signal_axes[0].units == "px"
        assert s.axes_manager.signal_axes[1].units == "px"
